// https://github.com/roe-dl/weathericons/tree/master
const axios = false; // require("axios");
const unzipper = require("unzipper");
const sax = require("sax");
const request = require('request');
const https = require('node:https');
const http = require("node:http");
const { scheduleTask } = require("cronosjs");
const MosmixElement = require("../data/MosmixElement.json");
const MosmixIcon = require("../data/MosmixIcon.json");
const MosmixStation = require("../data/MosmixStation.json")
const UNIX_MUL = 1000;

// var cluster = require('cluster');
// var numCPUs = require('os').cpus().length;

function MosmixStation_get(Id) {
    const obj = MosmixStation[Id];
    return { "name": obj[0], "lat": obj[1], "lon": obj[2], "alt": obj[3], "timeZone": obj[4] };
}

// const { Logg } = require('./redhm_cl_gl.js');
class Cl_Logger {
    constructor() {
    }
    setRed(red) {
        this.Log = red.log;
        Logg = this;

    }
    checkNode(_type, ...args) {
        if (typeof args[0] === 'object') {
            var func = args.shift();
            var checkType = _type
            if (_type === 'info')
                checkType = "log";
            var hh = func[checkType]; // prototype.warn 
            if (typeof hh === 'function') {
                func[checkType](args.join(' ').slice(0, 300));
                return;
            }
            if (typeof args[0] === 'object')
                args = [JSON.stringify(args[0])];
        }
        this.Log[_type](args.join(' ').slice(0, 300));
    }

    log(...args) { this.checkNode('trace', ...args); }
    trace(...args) { this.checkNode('trace', ...args); }
    debug(...args) { this.checkNode('debug', ...args); }
    info(...args) { this.checkNode('info', ...args); }
    warn(...args) { this.checkNode('warn', ...args); }
    error(...args) { this.checkNode('error', ...args); }
}

const Logg = new Cl_Logger();
const RwiX = require('./redhm-weather-icon.js');
const Rwi = RwiX.cl;

var geo_tz = null;
try {
    geo_tz = require('geo-tz')
    // do stuff
} catch (ex) {
    // console.log(ex);
}

class C {
    static INFOTXTSIZE = 16;
    static debug = false;
    static Hour12 = false;
}

let RED = null;

let Language = 'de'; // Intl.DateTimeFormat().resolvedOptions().locale.split("-")[0];
// Language = 'en';
function SetLanguage(la) {
    Language = la;
}

class Mosmix_Cl {
    static ElementsBase = ['TTT', 'Td', 'FF', 'DD', 'PPPP', 'FX1', 'ww', 'RR1c', 'RRS1c', 'N', 'VV']; // nicht in Mosmix_S: 'wwP', 'RRL1c',
    // wwP: "precipitation_perc", RRL1c: "rain",
    static altName = {
        TTT: "temp", Td: "dew_point", FF: "wind_speed", DD: "wind_deg", RR1c: "rain", RRS1c: "snow", PPPP: "pressure", FX1: "wind_gust", N: "clouds", VV: "visibility",
        wwP: "precipitation_perc"
    };
    static Elements_S = ["PPPP", "TX", "TTT", "Td", "TN", "T5cm", "DD", "FF", "FX1", "FX3", "FXh", "FXh25", "FXh40", "FXh55", "N", "Neff", "Nh", "Nm", "Nl", "N05", "VV", "wwM", "wwM6", "wwMh",
        "ww", "W1W2", "RR1c", "RRS1c", "RR3c", "RRS3c", "R602", "R650", "Rh00", "Rh02", "Rh10", "Rh50", "Rd02", "Rd50", "Rad1h", "SunD1",
        "wwP",
    ];

    getAllElement() {
        for (const Id of Object.keys(MosmixElement)) {
            const Element = MosmixElement[Id];
            if (Mosmix_Cl.altName[Id])
                Element.IdOut = Mosmix_Cl.altName[Id];
            else
                delete Element.IdOut;
        }
        /*
        for (const Id of Object.keys(MosmixIcon.icon)) {
            const Element = MosmixIcon.icon[Id];
                                    let Stufe = GetStufe(Number(Id));
                                    let keys = Object.keys(Stufe);
                                    if (keys.length > 0)
                                        Element.Stufe = Stufe;
        };*/
        return MosmixElement;
    }

    getElement(ele) {
        const obj = MosmixElement[ele];
        if (!obj) {
            Logg.error("getElement: " + ele);
            return "";
        }
        return obj;
    }
    getIdOut(ele) {
        const obj = this.getElement(ele);
        if (!obj || !obj.IdOut)
            return ele;
        return obj.IdOut;
    }
    getNoInterpol(ele) {
        const obj = this.getElement(ele);
        if (!obj || !obj.noIp)
            return false;
        return obj.noIp;
    }

    getDesc(ele) {
        const obj = this.getElement(ele);
        if (!obj)
            return "";
        if (Language == 'de')
            if (obj.de)
                return obj.de;
        return obj.en;
    }
    getUnit(ele) {
        const obj = this.getElement(ele);
        if (!obj)
            return "";
        return obj.Unit;
    }

    GetWW_Icon(ww) {
        const icon = MosmixIcon.icon[ww];
        if (!icon) {
            Logg.error("GetIconFromWW: " + ww);
            return;
        }
        return icon;
    }

    GetWW_Desc(ww) {
        const icon = this.GetWW_Icon(ww);
        if (!icon)
            return {};
        if (Language == 'de')
            if (icon.descD)
                return icon.descD;
        if (!icon.desc)
            return "WW: " + ww;
        return icon.desc;
    }

    GetWW_Stufe(ww) {
        const icon = this.GetWW_Icon(ww);
        if (!icon)
            return {};
        let Stufe = icon.Stufe;
        if (!Stufe)
            return {};
        return Stufe;
    }

    static findStation(lat, lon, best, callback = null) {
        let arrListNear = [];
        for (const id of Object.keys(MosmixStation)) {
            let obj = MosmixStation_get(id);
            let dist = Math.round(getDistanceFromLatLonInKm(lat, lon, obj.lat, obj.lon) * 10) / 10;
            arrListNear.push({ id: id, dist: dist });
        };
        arrListNear.sort((a, b) => {
            if (a.dist == b.dist)
                return a.id - b.id;
            return Number(a.dist) - Number(b.dist);
        });

        let anz = 10;
        if (best > 0)
            anz = best;
        let objList = {};
        let arrListRed = [];
        for (let index = 0; index < anz; index++) {
            const id = arrListNear[index].id;
            const dist = arrListNear[index].dist;
            objList[id] = MosmixStation_get(arrListNear[index].id);
            objList[id].dist = dist;
            let txt = "";
            /*            for (let index = 0; index < 5 - dist.toFixed(0).length; index++) {
                            txt += " ";
                        }*/
            txt += dist.toFixed(1) + " km " + id + ": " + objList[id].name + " " + objList[id].lat + " " + objList[id].lon + " " + objList[id].alt;
            arrListRed.push({ value: id, label: txt })

        }

        if (callback) {
            const topic = "worldmap";
            callback({
                topic: topic,
                payload: {
                    "layer": "MosmixStation", "name": "Haus", "icon": ":house:", "lat": lat, "lon": lon,
                    command: { "lat": lat, "lon": lon, zoom: 10 }
                }
            });
            for (const id of Object.keys(objList)) {
                const obj = objList[id];
                let out = { "layer": "MosmixStation", "name": obj.name, "Id": id, "icon": "fa-sun-o", "lat": obj.lat, "lon": obj.lon, alt: obj.alt, "dist": obj.dist };
                callback({ topic: topic, payload: out });
            }

            // let json = JSON.stringify(arrShort);
            // let jsonZip = zlib.gzipSync(json, { level: zlib.constants.Z_BEST_COMPRESSION });
            callback({ topic: "MosmixStationFind", payload: objList, kml: makeKml(objList) });
        }
        return arrListRed;
    }

}
let Mosmix = new Mosmix_Cl;

function getDpsInit() {
    return {
        humidity: { "unit": "%", "desc": (Language == 'de') ? "relative Luftfeuchte" : "relative humidity" },
        precipitation: { "unit": "mm", "desc": (Language == 'de') ? "Niederschlag" : "precipitation" },
    };
}

function getIsPointInsidePolygon(point, vertices) {
    const x = point[0]
    const y = point[1]

    let inside = false
    for (let i = 0, j = vertices.length - 1; i < vertices.length; j = i++) {
        const xi = vertices[i][0],
            yi = vertices[i][1]
        const xj = vertices[j][0],
            yj = vertices[j][1]

        const intersect = yi > y != yj > y && x < ((xj - xi) * (y - yi)) / (yj - yi) + xi
        if (intersect) inside = !inside
    }

    return inside
}

let WarnCellId = { Arr: [] };
function WarnCellId_Get(pos) {
    // console.log(pos);
    if (WarnCellId.Arr.length)
        return Promise.resolve(WarnCellId)

    let UsePos = null;
    if (pos) {
        try {
            UsePos = JSON.parse(pos);
        }
        catch (e) {
            Logg.error('Unexpected response: ' /*+ data*/);
        }
    }

    return new Promise((resolve, reject) => {
        https.get('https://maps.dwd.de/geoserver/wfs?SERVICE=WFS&VERSION=2.0.0&REQUEST=GetFeature&TYPENAMES=Warngebiete_Kreise&OUTPUTFORMAT=json', (res) => {
            // Logg.info ( this, 'statusCode:', res.statusCode);
            if (res.statusCode != 200) {
                reject("statusCode: " + res.statusCode);
                return;
            }
            // console.log('headers:', res.headers);
            let data = "";

            res.on('data', (chunk) => {
                // console.log(chunk);
                data += chunk;
            });
            res.on('end', () => {
                try {
                    let json = JSON.parse(data);
                    // console.log(json);
                    let region = {};

                    let arr = json.features;

                    // debugger;
                    WarnCellId.defId = "";
                    arr.forEach((obj, i) => {
                        let properties = obj.properties;
                        const name = properties.BL + " " + properties.KURZNAME;
                        region[name] = properties.WARNCELLID;
                        if (UsePos && UsePos.lon && UsePos.lat)
                            obj.geometry.coordinates.forEach((poly) => {
                                if (getIsPointInsidePolygon([UsePos.lon, UsePos.lat], poly)) {
                                    WarnCellId.defId = properties.WARNCELLID;
                                    WarnCellId.defName = name;
                                    // console.log(name);
                                }
                            });
                    });

                    let list = {};
                    for (const it of Object.keys(region).sort()) {
                        list[it] = region[it];
                    }
                    // console.log(list);
                    // console.log(data);
                    WarnCellId.Arr = [];
                    WarnCellId.Arr.push({ value: "", label: "k.A." })
                    for (const it of Object.keys(list)) {
                        WarnCellId.Arr.push({ value: String(list[it]), label: it })
                    }
                    resolve(WarnCellId);
                } catch (e) {
                    Logg.error('Unexpected response: ' /*+ data*/);
                    reject("try err");
                }
            });

        }).on('error', (e) => {
            Logg.error(e);
            reject("Err");
        });
    });
}

class wf_cl {
    M_Ele = [];  // Array with MOSMIX elements to process;

    static MOSMIX_L_URL = 'https://opendata.dwd.de/weather/local_forecasts/mos/MOSMIX_L/single_stations/{$station}/kml/MOSMIX_L_LATEST_{$station}.kmz';
    // static MOSMIX_ALL_URL = 'https://opendata.dwd.de/weather/local_forecasts/mos/MOSMIX_L/all_stations/kml/MOSMIX_L_LATEST.kmz';
    static MOSMIX_S_URL = 'https://opendata.dwd.de/weather/local_forecasts/mos/MOSMIX_S/all_stations/kml/MOSMIX_S_LATEST_240.kmz';
    static MOSMIX_MAXAGE = ((30 * 60)) * 1000; // um h:25 grobe Veröffentlichung, 

    constructor() {
        this.StMinMax = {};
        this.wfObj = {}; // main data structure to hold weather forecast. See initWeatherForecast()
    }
    init(mosmixStation) {
        if (!mosmixStation || mosmixStation == {})
            Logg.log("LEER");
        this.wfObj[mosmixStation] = {
            "description": "",
            nextUpdate_L: 0,
            nextUpdate_S: 0,
            dp: {}
            // followed by additional fields like 'TTT': Array of forecast values
        };
    }
    getInterpolatedValue(mosmixStation, attribute, forecastDateNr) {
        var idx = this.getTimeIndex(mosmixStation, forecastDateNr);

        this.assertAttributeExists(mosmixStation, attribute);
        let wfa = this.wfObj[mosmixStation].dp[attribute];
        let wft = this.wfObj[mosmixStation].times;

        if (idx == 0) {
            // all predictions in file are for the future => return first one
            return wfa[0];
        } else if (Number.isNaN(wfa[idx - 1])) {
            // non-continuous field, so no interpolation possible
            return wfa[idx];
        } else {
            // linear interpolation of current temperature
            var share = (forecastDateNr - wft[idx - 1]) / (wft[idx] - wft[idx - 1]);
            return wfa[idx - 1] + share * (wfa[idx] - wfa[idx - 1]);
        }
    }
    sumFutureValue(mosmixStation, attribute, hours, forecastDateNr) {
        var idx = this.getTimeIndex(mosmixStation, forecastDateNr);

        this.assertAttributeExists(mosmixStation, attribute);
        let wfa = this.wfObj[mosmixStation].dp[attribute];

        var sum = 0;
        // sum x future values (x = hours), but not more than length of array
        for (var i = idx; i < wfa.length && i < hours + idx; i++) {
            if (!isNaN(wfa[i])) {
                sum = sum + wfa[i];
            }
        }
        return sum;
    }
    getTimeIndex(mosmixStation, forecastDateNr) {
        if (!this.wfObj[mosmixStation]) {
            throw new Error(RED._("redhm-weather.warn.noDataForStation") + " " + mosmixStation);
        }
        var idx = this.wfObj[mosmixStation].times.findIndex((myDate) => {
            return (myDate > forecastDateNr);
        });
        if (idx == -1) {
            // no predictions for any future dates found - likely the file is too old
            throw new Error(RED._("redhm-weather.warn.nofuturepredictions") + " " + mosmixStation);
        }
        return idx;
    }

    assertAttributeExists(mosmixStation, attribute) {
        if (!this.wfObj[mosmixStation].dp[attribute]) {
            // attribute has not been parsed
            throw new Error(RED._("redhm-weather.warn.noattribute", { attribute }));
        }
    }

    update_L(mStation) {
        this.init(mStation);

        this.wfObj[mStation].nextUpdate_L = (new Date).getTime() + 600 * 1000; // retry in 10 minutes in case of a failure

        let xmlTagStack = [];
        let xmlStreamParser = sax.createStream(true, {
            'trim': true
        });
        let wf = this.wfObj[mStation];
        wf.times = [];
        xmlStreamParser.onopentag = (tag) => {
            switch (tag.name) {
                case "kml:name":
                    break;
                case "kml:Placemark":
                    break;
                default:
                    break;
            }
            if (!tag.isSelfClosing) {
                xmlTagStack.push(tag);
            }
        };
        xmlStreamParser.onclosetag = (tag) => {
            switch (tag) {
                case "kml:Placemark":
                    break;
                case "kml:name":
                    break;

                default:
                    break;
            }
            xmlTagStack.pop();
        };

        xmlStreamParser.ontext = (text) => {
            if (xmlTagStack.length) {
                var currentTag = xmlTagStack[xmlTagStack.length - 1];

                switch (currentTag.name) {
                    case "dwd:value":
                        if (xmlTagStack.length >= 2) {
                            var enclosingTag = xmlTagStack[xmlTagStack.length - 2];
                            if (enclosingTag.name == "dwd:Forecast") {
                                const eleName = enclosingTag.attributes["dwd:elementName"];
                                if (eleName) {
                                    if (this.M_Ele.includes(eleName)) {
                                        wf.dp[eleName] = text.split(/\s+/).map(v => Number.parseFloat(v));
                                    }
                                }
                            }
                        }
                        break;

                    case "dwd:TimeStep":
                        wf.times.push(new Date(text).getTime());
                        break;

                    case "kml:description":
                        wf.description = text;
                        break;

                    case "kml:coordinates":
                        wf.coordinates = text;
                        break;

                    case "kml:name":
                        // wf = this.wfObj[text];
                        if (text != mStation)
                            console.log("Name?: " + currentTag.name + ": " + text + " != " + mStation);
                        break;

                    case "dwd:IssueTime":
                        wf.IssueTime = new Date(text).getTime();
                        // console.log(new Date(wf.IssueTime).toLocaleString());

                        break;

                    default:
                        // case "dwd:Issuer":
                        // case "dwd:ProductID":
                        // case "dwd:GeneratingProcess":
                        // case "dwd:IssueTime":
                        // case "dwd:DefaultUndefSign":
                        break;
                }
            }
        };

        return new Promise((resolve, reject) => {
            //console.log(C.MOSMIX_URL.replace(/\{\$station\}/g, mStation));
            function Ok() {
                // 4x täglich (um 4, 10, 16 und 22 UTC) aktualisiert.
                const tOff = new Date().getTimezoneOffset() / 60;
                /*            let hAr = [4, 10, 16, 22];
                            const ist = new Date().getHours() + tOff;
                            let ff = hAr.find((element) => element > ist);
                            if (!ff)
                                ff = hAr[0];
                            let dd = new Date();
                            dd.setMinutes(10);
                            dd.setHours(ff - tOff);
                            this.wfObj[mStation].nextUpdate_L = dd.getTime();
                            console.log(new Date(this.wfObj[mStation].nextUpdate_L).toLocaleString());
                            // console.log(new Date(dd.getTime()).toLocaleString());*/
                this.wfObj[mStation].nextUpdate_L = new Date(wf.IssueTime + ((6 - tOff) * 60 * 60 + 10 * 60) * 1000).getTime();
                resolve();
            }
            unzipper.Open.url(request, wf_cl.MOSMIX_L_URL.replace(/\{\$station\}/g, mStation)).then((response) => {
                response.files[0].stream()
                    .on('error', reject)
                    .pipe(xmlStreamParser)
                    .on('error', reject)
                    .on('end', Ok.bind(this));
            }).catch((error) => {
                if (error.response && error.response.status == 404) {
                    reject(RED._("redhm-weather.warn.noDataForStation"), { ID: mStation });
                } else {
                    let message = error.response ? error.response.status + " " + error.response.statusText : 'Response is null';
                    reject(message);
                    //reject(response.status + " " + response.statusText);
                };
            });
            return;

            if (!axios) {
                https.get(wf_cl.MOSMIX_L_URL.replace(/\{\$station\}/g, mStation),
                    (res) => {
                        // console.log("headers:", res.headers);
                        res.pipe(unzipper.ParseOne(/\.kml/i)).on('error', reject)
                            .pipe(xmlStreamParser)
                            .on('error', reject)
                            .on('end', Ok.bind(this));

                        /*                     res.on("data", (chunk) => {
                      // console.log(chunk);
                                             });*/
                    }).on("error", (error) => {
                        if (error.response && error.response.status == 404) {
                            reject(RED._("redhm-weather.warn.noDataForStation"), { ID: "mStation" });
                        } else {
                            let message = error.response ? error.response.status + " " + error.response.statusText : 'Response is null';
                            reject(message);
                            //reject(response.status + " " + response.statusText);
                        };
                    });
                return;
            }
            axios({
                method: "get",
                url: wf_cl.MOSMIX_L_URL.replace(/\{\$station\}/g, mStation),
                responseType: "stream"
            }).then((response) => {
                response.data.pipe(unzipper.ParseOne(/\.kml/i))
                    .on('error', reject)
                    .pipe(xmlStreamParser)
                    .on('error', reject)
                    .on('end', Ok.bind(this));
            }).catch((error) => {
                if (error.response && error.response.status == 404) {
                    reject(RED._("redhm-weather.warn.noDataForStation"), { ID: "mStation" });
                } else {
                    let message = error.response ? error.response.status + " " + error.response.statusText : 'Response is null';
                    reject(message);
                    //reject(response.status + " " + response.statusText);
                };
            });
        });
    }

    update_S(mStationAr, node) {
        for (const st of Object.keys(this.wfObj)) {
            if (!mStationAr.includes(st))
                mStationAr.push(st);
        }
        mStationAr.forEach(mStation => {
            if (!mStation)
                return;
            this.init(mStation);
            this.wfObj[mStation].nextUpdate_S = (new Date).getTime() + 600 * 1000; // retry in 10 minutes in case of a failure
            this.wfObj[mStation].times = [];
        });
        this.StMinMax = { min: { station: "", val: Number.POSITIVE_INFINITY }, max: { station: "", val: Number.NEGATIVE_INFINITY } };

        if (node)
            node.outSpezial({ topic: "worldmap", payload: { "command": { "clear": "Mosmix", zoom: 0 } } });

        // const stations = Object.keys(this.wfObj);
        let arrList = [];
        let eleList = [];

        let ok = false;
        let xmlTagStack = [];
        let IssueTime;
        let xmlStreamParser = sax.createStream(true, {
            'trim': true
        });
        let AktStation = "";
        let wf = null;
        let times = [];
        let idx = 0;
        let place = null;
        xmlStreamParser.onopentag = (tag) => {
            switch (tag.name) {
                case "kml:name":
                    ok = false;
                    break;
                case "kml:Placemark":
                    place = {};
                    break;
                default:
                    break;
            }
            if (!tag.isSelfClosing) {
                xmlTagStack.push(tag);
            }
        };
        xmlStreamParser.onclosetag = (tag) => {
            switch (tag) {
                case "kml:Placemark":
                    arrList.push(place);
                    if (node) {
                        if (place.ww > 3) {
                            const a = Mosmix.GetWW_Icon(place.ww);
                            if (a && a.svg) {
                                // msg.payload = `<svg xmlns="http://www.w3.org/2000/svg" version="1.1" width="90%" height="90%" viewBox="-25 -25 50 50" transform="translate(2,2)">${a.svg}</svg>`;
                                const icon = `<svg xmlns="http://www.w3.org/2000/svg" version="1.1" width="90%" height="90%" viewBox="-25 -25 50 50" transform="translate(2,2)">${a.svg}</svg>`;
                                const data = "data:image/svg+xml;base64," + btoa(icon);
                                const info = Mosmix.GetWW_Desc(place.ww);
                                node.outSpezial({ topic: "worldmap", payload: { "layer": "Mosmix", "tooltip": place.name + " " + info, "name": place.name, Info: info, Id: place.Id, "icon": data, "lat": place.lat, "lon": place.lon } });
                            }
                        }
                    }
                    break;
                case "kml:name":
                    if (ok) {
                        // this._parser.resolve()
                    }
                    break;

                default:
                    break;
            }
            xmlTagStack.pop();
        };

        xmlStreamParser.ontext = (text) => {
            if (xmlTagStack.length) {
                var currentTag = xmlTagStack[xmlTagStack.length - 1];

                switch (currentTag.name) {
                    case "dwd:value":
                        if (xmlTagStack.length >= 2) {
                            var enclosingTag = xmlTagStack[xmlTagStack.length - 2];
                            if (enclosingTag.name == "dwd:Forecast") {
                                const eleName = enclosingTag.attributes["dwd:elementName"];
                                if (!eleList.includes(eleName))
                                    eleList.push(eleName);

                                if (ok) {
                                    if (eleName) {
                                        if (this.M_Ele.includes(eleName)) {
                                            wf.dp[eleName] = text.split(/\s+/).map(v => Number.parseFloat(v));
                                        }
                                    }
                                }
                                if (eleName == "ww") {
                                    const werte = text.split(/\s+/).map(v => Number.parseFloat(v));
                                    place.ww = werte[idx];
                                }
                                if (eleName == "TTT") {
                                    let mm = getMinMax(text.split(/\s+/).map(v => Number.parseFloat(v)));
                                    let min = mm.min;
                                    if (min < this.StMinMax.min.val) {
                                        this.StMinMax.min.val = min;
                                        this.StMinMax.min.station = AktStation;
                                        // console.log("min: " + this.StMinMax.min.station + " " + getTempCelsius(this.StMinMax.min.val).toFixed(1));
                                    }
                                    let max = mm.max;
                                    if (max > this.StMinMax.max.val) {
                                        this.StMinMax.max.val = max;
                                        this.StMinMax.max.station = AktStation;
                                        // console.log("max: " + this.StMinMax.max.station + " " + getTempCelsius(this.StMinMax.max.val).toFixed(1));
                                    }
                                }
                            }
                        }
                        break;

                    case "dwd:TimeStep":
                        times.push(new Date(text).getTime());
                        idx = times.findIndex((myDate) => {
                            return (myDate > new Date().getTime());
                        });

                        break;

                    case "kml:description":
                        place.name = text;
                        if (ok)
                            wf.description = text;
                        break;

                    case "kml:coordinates":
                        const [lon, lat, alt] = text.split(",");
                        place.lat = Number(lat);
                        place.lon = Number(lon)
                        place.alt = Math.round(Number(alt));
                        if (ok)
                            wf.coordinates = text;
                        break;

                    case "kml:name":
                        place.Id = text;
                        ok = false;
                        AktStation = text;
                        if (mStationAr.includes(text)) {
                            place.id = text;
                            ok = true;
                            wf = this.wfObj[text];
                            wf.times = times;
                            if (node)
                                node.status({ fill: "green", shape: "dot", text: "Station: " + text });
                        }
                        /*                    if (text != mStation)
                                            console.log("Name?: " + currentTag.name + ": " + text + " != " + mStation);**/
                        break;

                    case "dwd:IssueTime":
                        IssueTime = new Date(text).getTime();
                        // console.log(new Date(IssueTime).toLocaleString());
                        break;

                    default:
                        // case "dwd:Issuer":
                        // case "dwd:ProductID":
                        // case "dwd:GeneratingProcess":
                        // case "dwd:IssueTime":
                        // case "dwd:DefaultUndefSign":
                        break;
                }
            }
        };

        return new Promise((resolve, reject) => {
            //console.log(C.MOSMIX_URL.replace(/\{\$station\}/g, mStation));
            const start = new Date().getTime();
            function Ok() {
                console.log(((new Date().getTime() - start) / 1000).toFixed(0), "sec");
                {
                    arrList.sort((a, b) => {
                        if (a.lat == b.lat) {
                            if (a.lon == b.lon) {
                                if (a.alt == b.alt) {
                                    // console.log("doppelt:", JSON.stringify(a), JSON.stringify(b));
                                    if (a.Id < b.Id)
                                        return -1;
                                    if (a.Id > b.Id)
                                        return 1;
                                    // names must be equal
                                    return 0;
                                }
                                return a.alt - b.alt;
                            }
                            return a.lon - b.lon;
                        }
                        return a.lat - b.lat;
                    }
                    );

                    let tz = {};

                    class getMinMax {
                        constructor(objList, ele) {
                            this.min = { wert: Number.POSITIVE_INFINITY };
                            this.max = { wert: Number.NEGATIVE_INFINITY };
                            this.objList = objList;
                            this.ele = ele;
                            return this;
                        }
                        get() {
                            for (const id of Object.keys(this.objList)) {
                                const obj = this.objList[id];
                                const wert = obj[this.ele];
                                if (Number.isNaN(wert))
                                    continue;
                                if (wert > this.max.wert) {
                                    this.max.wert = wert;
                                    this.max.id = id;
                                }
                                if (wert < this.min.wert) {
                                    this.min.wert = wert;
                                    this.min.id = id;
                                }
                            }
                            return this;
                        }
                    }
                    let objList = {};
                    arrList.forEach(function (obj) {
                        // ["91753","HIHIFO",-13.23,-176.17,23]

                        let timeZone = "";
                        if (geo_tz) {
                            let timeZ = geo_tz.find(obj.lat, obj.lon);
                            if (!timeZ.length)
                                console.log("Err", obj.lat, obj.lon);
                            else {
                                timeZone = timeZ[0];
                                if (false)
                                    if (timeZ.length > 1) {
                                        console.log("Warn: Mehr als eine");
                                        timeZ.forEach(timeZone => {
                                            console.log(obj.Id, timeZone, timeZoneNameShort(timeZone), timeShort(new Date() / UNIX_MUL, timeZone), obj.lat, obj.lon);
                                        });
                                    }
                            }
                            if (!tz[timeZone])
                                tz[timeZone] = 0;
                            tz[timeZone]++;
                        }
                        // Änderung MosmixStation_get anpassen
                        objList[obj.Id] = [obj.name, obj.lat, obj.lon, obj.alt, timeZone];
                    });
                    let alt = new getMinMax(objList, 3).get();
                    let lat = new getMinMax(objList, 1).get();
                    if (node) {
                        node.outSpezial({ topic: "worldmap", payload: { "command": { zoom: 10 } } });
                        node.outSpezial({ topic: "MosmixStationList", payload: objList, kml: makeKml(arrList), eleList: eleList, mix: { tz: tz, alt: alt, lat: lat } });
                    }

                }
                for (const mStation of Object.keys(this.wfObj)) {
                    let wf = this.wfObj[mStation];
                    wf.IssueTime = IssueTime;
                    let nexDate = new Date(wf.IssueTime);
                    nexDate.setMinutes(30);
                    nexDate.setSeconds(0);
                    let add = 0;
                    if (nexDate <= new Date())
                        add = 60 * 60 * 1000;

                    wf.nextUpdate_S = nexDate.getTime() + add;

                    const tOff = new Date().getTimezoneOffset() / 60;
                    wf.nextUpdate_L = new Date(wf.IssueTime + ((6 - tOff) * 60 * 60 + 10 * 60) * 1000).getTime();

                    // console.log("Next Update: " + new Date(wf.nextUpdate_S).toLocaleString() + " " + new Date().toLocaleString() + " "
                    // + ((new Date(wf.nextUpdate_S) - new Date().getTime()) / (1000 * 60)).toFixed(1) + " min " + new Date(wf.IssueTime).toLocaleString());
                    // wf.nextUpdate_S = new Date(wf.times[0]).getTime() + wf_cl.MOSMIX_MAXAGE;
                };
                resolve();
            }

            unzipper.Open.url(request, { url: wf_cl.MOSMIX_S_URL, concurrency: 1 }).then((response) => {
                response.files[0].stream()
                    .on('error', reject)
                    .pipe(xmlStreamParser)
                    .on('error', reject)
                    .on('end', Ok.bind(this));
            }).catch((error) => {
                if (error.response && error.response.status == 404) {
                    reject(RED._("redhm-weather.warn.noDataForStation"), { ID: "mStation" });
                } else {
                    let message = error.response ? error.response.status + " " + error.response.statusText : 'Response is null';
                    reject(message);
                    //reject(response.status + " " + response.statusText);
                };
            });
            return;

            if (!axios) {
                let options = { highWaterMark: 512 * 1024 };
                http.get(wf_cl.MOSMIX_S_URL, options,
                    (res) => {
                        // console.log("headers:", res.headers);
                        res.pipe(unzipper.ParseOne(/\.kml/i))
                            .on('error', reject)
                            .pipe(xmlStreamParser)
                            .on('error', reject)
                            .on('end', Ok.bind(this));
                    }).on("error", (error) => {
                        if (error.response && error.response.status == 404) {
                            reject(RED._("redhm-weather.warn.noDataForStation"), { ID: "mStation" });
                        } else {
                            let message = error.response ? error.response.status + " " + error.response.statusText : 'Response is null';
                            reject(message);
                        };
                    });
                return;
            }
            axios({
                method: "get",
                url: wf_cl.MOSMIX_S_URL,
                responseType: "stream"
            }).then((response) => {
                response.data.pipe(unzipper.ParseOne(/\.kml/i)) // ,{forceStream: false, concurrency: 4}
                    .on('error', reject)
                    .pipe(xmlStreamParser)
                    .on('error', reject)
                    .on('end', Ok.bind(this));
            }).catch((error) => {
                if (error.response && error.response.status == 404) {
                    reject(RED._("redhm-weather.warn.noDataForStation"), { ID: "mStation" });
                } else {
                    let message = error.response ? error.response.status + " " + error.response.statusText : 'Response is null';
                    reject(message);
                    //reject(response.status + " " + response.statusText);
                };
            });
        });
    }

    get() {
        return this.wfObj;
    }

    getInfo() {
        let Stat = {};
        for (const id of Object.keys(this.wfObj).sort()) {
            Stat[id] = this.wfObj[id].description;
        }
        return { MinMax: this.StMinMax, Stations: Stat };
    }

    updateIfOutdated(Mosmix_S, mStatAr, that) {
        if (!Mosmix_S) {
            if (C.debug) return this.update_L(mStatAr);

            let mStation = mStatAr[0];
            let wf = this.wfObj[mStation];
            if (!wf || !wf.nextUpdate_L || (new Date().getTime() > wf.nextUpdate_L)) {
                that.status({ fill: "blue", shape: "dot", text: RED._("redhm-weather.status.requesting", { ID: mStation }) + " " + new Date().toLocaleString(Language) });
                return this.update_L(mStation);
            } else {
                that.status({ fill: "green", shape: "dot", text: "Update in " + ((new Date(wf.nextUpdate_L).getTime() - new Date().getTime()) / (60 * 1000)).toFixed(0) + " min" });
                return Promise.resolve();
            }
        }
        else {
            let nichtDa = false;
            let update = false;
            let UpdateTxt = "";
            mStatAr.forEach(mStation => {
                let wf = this.wfObj[mStation];
                if (!wf) {
                    nichtDa = true;
                    return;
                }
                if (!wf.nextUpdate_S || (new Date().getTime() > wf.nextUpdate_S)) {
                    update = true;
                }
                else
                    UpdateTxt = ((new Date(wf.nextUpdate_S).getTime() - new Date().getTime()) / (60 * 1000)).toFixed(0) + " min";
            });

            if (nichtDa || update) {
                that.status({ fill: "blue", shape: "dot", text: RED._("redhm-weather.status.requesting", { ID: "" }) + " " + new Date().toLocaleString(Language) });
                return this.update_S(mStatAr, that);
            } else {
                that.status({ fill: "green", shape: "dot", text: "Update in " + UpdateTxt });
                return Promise.resolve("noUpdate");
            }
        }
    }

    getTemperature(mosmixStation, forecastDateNr) {
        return Math.round(getTempCelsius(this.getInterpolatedValue(mosmixStation, "TTT", forecastDateNr)) * 10) / 10;
    }
    getHumidity(mosmixStation, forecastDateNr) {
        // calculate relative humidity from Taupunkt and temp in Celsius
        return getRelHumidity(getTempCelsius(this.getInterpolatedValue(mosmixStation, "TTT", forecastDateNr)), getTempCelsius(this.getInterpolatedValue(mosmixStation, "Td", forecastDateNr)));
    }
}

let wFC = new wf_cl();

function getMinMax(arr) {
    let out = { min: Number.POSITIVE_INFINITY, max: Number.NEGATIVE_INFINITY };

    arr.forEach(wert => {
        if (Number.isNaN(wert))
            return;
        if (wert > out.max)
            out.max = wert;
        if (wert < out.min)
            out.min = wert;
    });
    return out;
}

const kelvin = 273.15;
function getTempCelsius(tempK) {
    return tempK - kelvin;
}

function getRelHumidity(Temp, TauTemp) {
    function getSDD(tempC) {
        // returns Sättigungsdampfdruck in hPa given a temperature in Celsius
        if (tempC >= 0) {
            return 6.1078 * Math.pow(10, (7.5 * tempC) / (237.3 + tempC));
        } else {
            return 6.1078 * Math.pow(10, (7.6 * tempC) / (240.7 + tempC));
        }
    }
    // calculate relative humidity from Taupunkt and temp in Celsius
    return Math.round(1000 * getSDD(TauTemp) / getSDD(Temp)) / 10;
}

function UnitGetOut(that, Unit) {
    switch (Unit) {
        case "kg/m2":
            return "mm";

        case 'K':
        case 'C':
            switch (that.UnitTemp) {
                case 'C':
                    return "°C";

                case 'F':
                    return "°F";

                default:
                    break;
            }
            break;

        case 'm/s':
            switch (that.UnitSpeed) {
                case 'km/h':
                    return "km/h";

                case 'mph':
                    return "mph";

                default:
                    break;
            }
            break;

        case 'Pa':
            return 'hPa';
        default:
            break;
    }
    return Unit;
}

function fixUnit2(that, istUnit, wert) {
    if (Number.isNaN(wert))
        return 0;

    switch (istUnit) {
        case 'K':
            switch (that.UnitTemp) {
                case 'C':
                    wert = getTempCelsius(wert);
                    break;

                case 'F':
                    wert = (wert - kelvin) * 9 / 5 + 32;
                    break;

                case 'K':
                default:
                    break;
            }
            break;

        case 'C':
            switch (that.UnitTemp) {
                case 'C':
                    break;

                case 'F':
                    wert = (wert * 9 / 5) + 32;
                    break;

                case 'K':
                    wert = wert + kelvin;
                    break;
                default:
                    break;
            }
            break;

        case 'm/s':
            switch (that.UnitSpeed) {
                case 'km/h':
                    wert = wert * 3.6;
                    break;

                case 'mph':
                    wert = wert * 2.2369362921;
                    break;

                default:
                    break;
            }
            break;

        case 'Pa':
            wert = wert / 100;
            break;

        default:
            break;
    }
    return Math.round(wert * 100) / 100;
}

function makeKml(objList) {
    let xml = `<?xml version="1.0" encoding="UTF-8"?><kml xmlns="http://www.opengis.net/kml/2.2"><Folder><name>Mosmix Stationen</name>`;
    for (const id of Object.keys(objList)) {
        let obj = objList[id];
        xml += `<Placemark><name>${id}</name><description>${obj.name}</description><Point><coordinates>${obj.lon},${obj.lat},${obj.alt}</coordinates></Point></Placemark>`;
    }
    xml += '</Folder></kml>';
    return xml;
}

function getDistanceFromLatLonInKm(lat1, lon1, lat2, lon2) {
    var R = 6371; // Radius of the earth in km
    var dLat = deg2rad(lat2 - lat1);
    var dLon = deg2rad(lon2 - lon1);
    var a =
        Math.sin(dLat / 2) * Math.sin(dLat / 2) +
        Math.cos(deg2rad(lat1)) * Math.cos(deg2rad(lat2)) *
        Math.sin(dLon / 2) * Math.sin(dLon / 2);
    var c = 2 * Math.atan2(Math.sqrt(a), Math.sqrt(1 - a));
    var d = R * c; // Distance in km
    return d;
}

function deg2rad(deg) {
    return deg * (Math.PI / 180)
}

function get_ww_text(ww) {
    if (Number.isNaN(ww))
        ww = 0;

    let pos = MosmixIcon.icon[ww];
    if (!pos) {
        pos = MosmixIcon.icon[0];
    }

    let text = "";
    if (Language == 'de')
        if (pos.descD)
            text = pos.descD;
    if (!text)
        text = pos.desc
    if (text != null)
        return text;
    // console.log(ww + "->" + text);
    return ww;
}

var StateList = null;
var StatePos = 0;
var StateAchtel_m8 = 0;

module.exports = function (RED_funk) {
    "use strict";
    RED = RED_funk;
    {
        const setLang = RED.settings.get("lang")
        if (setLang)
            SetLanguage(setLang);
    }
    if (false) {
        console.log(RED.settings.settingsFile);
        const setLang = RED.settings.get("test")
        if (!setLang)
            RED.settings.set("test", "läuft");
    }

    RED.httpAdmin.get('/RedDVD/:cmd', RED.auth.needsPermission('RedHM.read'), function (req, res) {
        //Logg.warn(this,req);
        //Logg.warn(this.global);
        //if ( GetGlobalContext() === this.global)
        // Logg.warn(this,clList());

        switch (req.params.cmd) {
            case "WarnCellId": {
                WarnCellId_Get(req.query.pos)
                    .then((dataIn) => {
                        res.status(200).send(JSON.stringify({ WarnCellId: dataIn }));
                        return;
                    })
                    .catch((err) => {
                        Logg.error(err);
                    })
                    .finally(() => {
                        // console.log('Experiment completed');
                    });
                return;
            }
            case "timeZone":
                {
                    let data = { arr: [], ElementsBase: Intl.DateTimeFormat().resolvedOptions().timeZone };
                    var aryIanaTimeZones = Intl.supportedValuesOf('timeZone');
                    /*                    var arrShort = [];
                                        aryIanaTimeZones.forEach((timeZone) => {
                                            let nameShort = timeZoneNameShort(timeZone);
                                            if (!arrShort.includes(nameShort))
                                                arrShort.push(nameShort);
                                        });*/
                    let date = new Date;
                    aryIanaTimeZones.forEach((timeZone) => {
                        let strTime = date.toLocaleString("en-US", { timeZone: `${timeZone}` });
                        let date2 = new Date(strTime);
                        let nameShort = timeZoneNameShort(timeZone);
                        const utc = (Math.round((date2.getTime() - date.getTime()) / (1000 * 60)) - new Date(strTime).getTimezoneOffset()) / 60;
                        // console.log(timeZone, strTime, utc  );
                        let add = "+";
                        if (utc < 0)
                            add = "";
                        // data.arr.push({ value: timeZone, label: "UTC: " + add + utc + " " + timeZone, sel: utc });
                        data.arr.push({ value: timeZone, label: nameShort + " " + timeZone, sel: utc });
                    });
                    data.arr.sort((a, b) => {
                        if (a.sel == b.sel) {
                            if (a.Id < b.Id)
                                return -1;
                            if (a.Id > b.Id)
                                return 1;
                            // names must be equal
                            return 0;
                        }
                        return Number(b.sel) - Number(a.sel);
                    });
                    res.status(200).send(JSON.stringify(data));
                }
                return;

            case "fieldList":
                {
                    let data = { arr: [], ElementsBase: wFC.M_Ele.toString() };

                    for (const Id of Object.keys(MosmixElement)) {
                        const obj = MosmixElement[Id];
                        // console.log(Id + " " + obj.IsBase + " " + obj.txt + " [" + obj.Unit + "]");
                        // if (!obj.IsBase)
                        {
                            let txt = "";
                            txt += Id;

                            if (obj.IdOut)
                                txt += "->" + obj.IdOut;
                            txt += " - " + obj.de;
                            if (obj.Unit)
                                txt += " [" + obj.Unit + "]";
                            if (obj.Range)
                                txt += " (" + obj.Range + ")";

                            if (!Mosmix_Cl.Elements_S.includes(Id)) {
                                if (wFC.M_Ele.includes(Id))
                                    console.log(txt);
                            }
                            else
                                data.arr.push({ value: Id, label: txt, sel: wFC.M_Ele.includes(Id) });
                        }
                        data.arr.sort((a, b) => {
                            if (a.sel == b.sel) {
                                if (a.Id < b.Id)
                                    return -1;
                                if (a.Id > b.Id)
                                    return 1;
                                // names must be equal
                                return 0;
                            }
                            return Number(b.sel) - Number(a.sel);
                        });
                    };
                    // console.log("->" + wFC.M_Ele.toString());
                    res.status(200).send(JSON.stringify(data));
                }
                return;

            case "getSvd":
                {
                    let icon = "Positionsangabe fehlt.";
                    // console.log(req.query.fill);
                    if (req.query.pos_config) {
                        /*
                        let results = [];
                        RED.nodes.eachNode(function(node) {
                           if (node.type.indexOf ("redhm-pos_config") >= 0) {
                              results.push(node);
                           }
                        });
                        let aa = RED.nodes.getCredentials(req.query.pos_config);
                        */
                        let posConf = RED.nodes.getNode(req.query.pos_config);
                        if (posConf && posConf.PosConfig) {
                            let pos = posConf.PosConfig;
                            if (pos && (pos.lat > 0) && (pos.lon > 0)) {
                                if (!StateList)
                                    StateList = RwiX.GetSvg_WetterTest();
                                // StateList = require("../data/WetterTest.json");
                                const Sun = new Sun_cl(new Date().getTime() / UNIX_MUL, pos.lat, pos.lon);
                                let Stufe = { ...StateList[StatePos], ...{ IsTag: Sun.IsTag, Moon: Moon_Cl.get(pos.lat).Moon, Achtel_m8: StateAchtel_m8 } };
                                // console.log ( Stufe);
                                StatePos++;
                                if (StatePos >= StateList.lenght)
                                    StatePos = 0;

                                StateAchtel_m8++;
                                if (StateAchtel_m8 > 8)
                                    StateAchtel_m8 = 0;
                                let icon1 = Rwi.GetSvg_Wetter(0, 0, 150, 150, Stufe, 'style="border: 1px solid black;background-color:white;"');
                                let icon2 = Rwi.GetSvg_Wetter(0, 0, 150, 150, Stufe, 'style="border: 1px solid black;background-color:grey;"');
                                icon = `
<div style="height: 100%; justify-content: center; align-items: center;">
<div>
    <div layout="rowicons" layout-align="space-around start" style="font-size:100%;padding: 0px;">
    ${icon1}
    ${icon2}
    </div>
</div>`;
                                // icon = "<div>" + Rwi.GetSvg_Wetter(0, 0, 150, 150, Stufe, 'style="background-color:white; border: 1px solid black; "');
                                // icon += Rwi.GetSvg_Wetter(0, 0, 150, 150, Stufe, 'style="background-color:grey; border: 1px solid black;"') + "</div>";
                                // console.log ( icon);
                            }
                        }
                        else
                            icon = '<i style="font-size:25px;color:red;"><b>DEPLOY<b></i>';
                    }
                    let data = { svg: icon };
                    res.status(200).send(JSON.stringify(data));
                }
                return;

            case "mosmixStation":
                {
                    // console.log(req.query);
                    let arr = [];
                    let posConf = RED.nodes.getNode(req.query.pos_config);
                    if (posConf && posConf.PosConfig && posConf.PosConfig) {
                        let pos = posConf.PosConfig;
                        if (pos && (pos.lat > 0) && (pos.lon > 0)) {
                            arr = Mosmix_Cl.findStation(Number(pos.lat), Number(pos.lon), 10);
                        }
                    }
                    res.status(200).send(JSON.stringify(arr));
                }
                return;

            case "pos_config":
                {
                    // console.log(req.query);
                    let data = "";
                    let posConf = RED.nodes.getNode(req.query.pos_config);
                    if (posConf && posConf.credentials) {
                        data = posConf.name;
                    }
                    res.status(200).send(JSON.stringify(data));
                }
                return;

            default:
                Logg.error("httpAdmin: " + req.params.cmd);
                res.status(500);
                return;
        }
    });

    class RedHM_DWD {
        constructor(nodeConfig) {
            RED.nodes.createNode(this, nodeConfig);
            this.pos_config = RED.nodes.getNode(nodeConfig.pos_config);
            if (!this.pos_config) {
                this.error(RED._('missing setup config'));
                return;
            }
            // updateList();
            // console.log ( RED._("redhm-weather.warn.noDataForStation"), { ID: "ALL" });
            // console.log ( RED._("node-red:common.label.name"));
            // console.log ( RED._("common.setup.lang"));
            // Language = RED._("redhm-weather.setup.lang");
            // console.log ( Language );
            // console.log (  RED.settings );
            // console.log (  RED.settings.get( "lang") );

            // console.log (  RED.settings.getUserSettings( "_") );
            // console.log (  RED.settings.settingsFile );

            this.defaultStation = this.credentials.mosmixStation;
            this.topic = nodeConfig.topic;
            wFC.M_Ele = Mosmix_Cl.ElementsBase.concat();
            let arr = nodeConfig.additionalFields.split(",")
            // console.log(config.additionalFields);
            arr.forEach(name => {
                if (!name)
                    return;
                if (!wFC.M_Ele.includes(name))
                    wFC.M_Ele.push(name);
            });
            wFC.M_Ele = wFC.M_Ele.sort();
            // console.log(wFC.M_Ele.toString());
            /*
                    this.interval_id = null;
                    if (this.repeat > 0) {
                        this.intervalId = setInterval(function () {
                            this.emit("input", {
                                payload: {}
                            });
                        }, this.repeat * 1000);
                    }
            */
            this.UnitTemp = nodeConfig.UnitTemp;
            this.UnitSpeed = nodeConfig.UnitSpeed;

            this.on('input', this.#input);
            this.on('close', this.#close);
            this.status({ fill: "yellow", shape: "ring", text: "ready" });
        };

        #fixUnit(id, wert) {
            return fixUnit2(this, Mosmix.getUnit(id), wert);
        }

        #input(msg, send, done) {
            let Mosmix_S = false;
            this.mStation
            switch (msg.command) {
                case "MosmixStationFind":
                    Mosmix_Cl.findStation(msg.payload.lat, msg.payload.lon, msg.payload.best, this.outSpezial.bind(this));
                    return;
                case "MosmixElement":
                    this.outSpezial({ topic: "MosmixElement", payload: Mosmix.getAllElement() });
                    this.outSpezial({ topic: "MosmixIcon", payload: MosmixIcon });
                    return;
                case "WetterTest":
                    this.outSpezial({ topic: "WetterTest", payload: RwiX.GetSvg_WetterTest() });
                    return;
                default:
                    break;
            }

            let pay = {};

            if (msg.mosmixStation) {
                this.mStation = [].concat(msg.mosmixStation);
                /*                if (msg.payload && msg.payload.timeZone)
                                    pay.timeZone = msg.payload.timeZone;
                                if (msg.payload && msg.payload.timezone_offset)
                                    pay.timezone_offset = msg.payload.timezone_offset;*/
            }
            else {
                this.mStation = [this.defaultStation];
            }

            let mStationAr = this.mStation;
            switch (msg.command) {
                case "MosmixS":
                    Mosmix_S = true;
                    mStationAr = [this.defaultStation];
                    for (let st of msg.payload.split(",")) {
                        st = st.trim();
                        if (st && st != "")
                            if (!mStationAr.includes(st))
                                mStationAr.push(st);
                    }
                    break;
            }

            if (this.prom && !this.prom.done) {
                this.warn("Transfer läuft...");
                return;
            }
            this.prom = wFC.updateIfOutdated(Mosmix_S, mStationAr, this);

            this.prom
                .then((aa) => {
                    if (Mosmix_S && (aa == "noUpdate"))
                        return;

                    if (this.topic)
                        msg.topic = this.topic;
                    let inTopic = msg.topic;

                    let fixUnit = this.#fixUnit.bind(this);

                    mStationAr.forEach(station => {
                        const wf = wFC.get()[station];
                        if (wf) {
                            if (false) {
                                let StMinMax = { min: { station: "", val: Number.POSITIVE_INFINITY }, max: { station: "", val: Number.NEGATIVE_INFINITY } };
                                let mm = getMinMax(wf.dp.TTT);
                                let min = mm.min;
                                if (min < StMinMax.min.val) {
                                    StMinMax.min.val = min;
                                    StMinMax.min.station = this.mStation;
                                    console.log("min: " + StMinMax.min.station + " " + StMinMax.min.val);
                                }
                                let max = mm.max;
                                if (max > StMinMax.max.val) {
                                    StMinMax.max.val = max;
                                    StMinMax.max.station = this.mStation;
                                    console.log("max: " + StMinMax.max.station + " " + StMinMax.max.val);
                                }
                            }

                            var forecastDate = new Date();
                            if (pay.lookAheadHours) {
                                var lookAheadHours = Number(pay.lookAheadHours) || 0;
                                forecastDate.setTime(forecastDate.getTime() + lookAheadHours * 3600000);
                            } else {
                                forecastDate.setTime(forecastDate.getTime());
                            }
                            const fdNr = forecastDate.getTime();
                            try {
                                if (!inTopic)
                                    msg.topic = wf.description;

                                let [lon, lat, alt] = wf.coordinates.split(",");
                                let idx = wFC.getTimeIndex(station, fdNr);
                                let hourly = [];
                                if (idx > 0)
                                    idx--;

                                for (let index = idx; index < idx + 48; index++) {
                                    const temp = getTempCelsius(wf.dp.TTT[index]);
                                    const dew_point = getTempCelsius(wf.dp.Td[index]);

                                    let hh = {
                                        dt: Math.round(wf.times[index] / 1000),
                                        humidity: getRelHumidity(temp, dew_point),
                                        description: get_ww_text(wf.dp.ww[index]),

                                    };
                                    function eleAdd(id) {
                                        if (!wFC.wfObj[station].dp[id])
                                            return;
                                        let name = Mosmix.getIdOut(id);
                                        //                                        wFC.assertAttributeExists(station, id);
                                        hh[name] = fixUnit(id, wf.dp[id][index]);
                                    }
                                    wFC.M_Ele.forEach(name => {
                                        eleAdd(name);
                                    });
                                    hh.precipitation = hh.rain + hh.snow;

                                    hourly.push(hh);
                                }

                                let days = {};
                                const aktTag = TagImJahr(new Date());
                                wf.times.forEach((ele, i) => {
                                    const dd = TagImJahr(new Date(ele)) - aktTag;
                                    if (!days[dd])
                                        days[dd] = [];
                                    days[dd].push(i);
                                });
                                let daily = [];
                                for (const dd of Object.keys(days)) {
                                    const arr = days[dd];
                                    if (arr.length == 24) {
                                        function getMinMax(id) {
                                            let dp = wf.dp[id];
                                            let min = Number.POSITIVE_INFINITY;
                                            let max = Number.NEGATIVE_INFINITY;
                                            let sum = 0;
                                            let anz = 0;
                                            arr.forEach((ele, i) => {
                                                const wert = dp[ele];
                                                if (Number.isNaN(wert))
                                                    return;
                                                anz++;
                                                if (min > wert)
                                                    min = wert;
                                                if (max < wert)
                                                    max = wert;
                                                sum += wert;
                                            });
                                            if (min == Number.POSITIVE_INFINITY)
                                                min = 0;
                                            if (max == Number.NEGATIVE_INFINITY)
                                                max = 0;
                                            min = fixUnit(id, min);
                                            max = fixUnit(id, max);
                                            let av = fixUnit(id, (sum / anz));
                                            sum = fixUnit(id, sum);
                                            return { min: min, max: max, sum: sum, av: av };
                                        }
                                        const dt = Math.round(wf.times[arr[0]] / 1000);

                                        const sr = new Sun_cl(dt, lat, lon);
                                        let dd = {
                                            "dt": dt, "sunrise": sr.sunrise, "sunset": sr.sunset
                                        };
                                        function eleAdd(id) {
                                            if (!wFC.wfObj[station].dp[id])
                                                return;
                                            let name = Mosmix.getIdOut(id);
                                            dd[name] = getMinMax(id);
                                        }
                                        wFC.M_Ele.forEach(name => {
                                            eleAdd(name);
                                        });
                                        const ww = dd.ww.max;
                                        if (ww)
                                            dd.description = get_ww_text(ww);
                                        dd.humidity = getRelHumidity(dd.temp.av, dd.dew_point.av);

                                        dd.ww = dd.ww.max;
                                        dd.dew_point = dd.dew_point.av;
                                        dd.pressure = dd.pressure.av;
                                        dd.visibility = dd.visibility.av;
                                        dd.clouds = dd.clouds.av;
                                        dd.wind_speed = dd.wind_speed.max;
                                        dd.wind_gust = dd.wind_gust.max;
                                        dd.wind_deg = dd.wind_deg.av;
                                        dd.rain = dd.rain.sum;
                                        dd.snow = dd.snow.sum;
                                        dd.precipitation = dd.rain + dd.snow;
                                        if (dd.precipitation_perc) // nicht Mosmix_S
                                            dd.precipitation_perc = dd.precipitation_perc.max;
                                        dd.Rad1h = dd.Rad1h.av; // uvi ?
                                        daily.push(dd);
                                    }
                                }

                                if (this.defaultStation == station) {
                                    lat = this.pos_config.PosConfig.lat;
                                    lon = this.pos_config.PosConfig.lon;
                                }
                                const humidity = wFC.getHumidity(station, fdNr);
                                const update = { Akt: Math.round(wf.IssueTime / UNIX_MUL), Next: Mosmix_S ? Math.round(wf.nextUpdate_S / UNIX_MUL) : Math.round(wf.nextUpdate_L / UNIX_MUL) };
                                msg.WeatherTyp = "DWD";
                                msg.NodeStation = (station == this.defaultStation) ? true : false;
                                let Bez = msg.WeatherTyp + " ";
                                if (msg.NodeStation && this.topic) Bez += this.topic;
                                else Bez += wf.description;

                                pay = {
                                    ...pay, ...{
                                        update: update,
                                        timeZone: MosmixStation_get(station).timeZone,
                                        Bez: Bez,
                                        lat: Number(lat), lon: Number(lon),
                                        current: {},
                                        hourly: hourly,
                                        daily: daily,
                                        // "station": wf.description,
                                        mosmixStation: { Id: station, name: wf.description, lat: Number(lat), lon: Number(lon), alt: Math.round(Number(alt)) },
                                        // "tempc": temp,
                                        // "humidity": humidity,
                                        // "windspeed": Math.round(ws * 10) / 10,
                                        // wind_speed: Math.round(ws * 3.6 * 100) / 100,
                                        // "winddirection": Math.round(wFC.getInterpolatedValue(station, "DD", fdNr) * 10) / 10,
                                        // "precipitation_perc": Math.round(wFC.getInterpolatedValue(station, "wwP", fdNr) * 10) / 10,
                                        "precipitationNext24h": Math.round(wFC.sumFutureValue(station, "RR1c", 24, fdNr) * 10) / 10,
                                        "forecast_dt": fdNr,
                                        "forecast_time": wf.times[idx],
                                        // txt: description,
                                        // ww: wf.dp.ww[idx],
                                        // idx: idx
                                    }
                                };
                                const wwTxt = get_ww_text(wf.dp.ww[idx]);
                                const dt = Math.round(fdNr / UNIX_MUL);

                                pay.current = {
                                    ...new Sun_cl(dt, lat, lon), ...{
                                        "dt": dt,
                                        // "feels_like": 5.56, 
                                        "humidity": humidity,
                                        // "uvi": 0, "clouds": 100, "visibility": 8000,
                                        // "weather": [{ "id": 0, "main": "", "description": "Leichter Regen", "icon": "" }],
                                        "description": wwTxt,
                                        // "ACTUAL_TEMPERATURE": Math.round(temp * 10) / 10,
                                        // "HUMIDITY": Math.round(humidity),
                                        // "wind_speed_mph": 4.63, "wind_gust_mph": 10.29, "wind_cardinal": "WNW" 
                                    }
                                };
                                // pay["precipitation%"] = pay.precipitation_perc; // for backward compatibility. Will be removed in the future
                                pay.Dps = getDpsInit();
                                wFC.M_Ele.forEach(id => {
                                    if (!wFC.wfObj[station].dp[id])
                                        return;
                                    let name = Mosmix.getIdOut(id);
                                    var val;
                                    if (Mosmix.getNoInterpol(id))
                                        val = wf.dp[id][idx];
                                    else
                                        val = wFC.getInterpolatedValue(station, id, fdNr);
                                    pay.Dps[name] = { unit: UnitGetOut(this, Mosmix.getUnit(id)), desc: Mosmix.getDesc(id) };
                                    if (val !== null) {
                                        pay.current[name] = fixUnit(id, val);
                                    };
                                });
                                pay.current.precipitation = pay.current.rain + pay.current.snow;

                                //msg.wi = MosmixIcon;
                                pay = addData(pay);
                                msg.payload = pay;
                                msg.Mosmix = wFC.getInfo();
                                // let iconHtml = `<svg xmlns="http://www.w3.org/2000/svg" version="1.1" width="${size}" height="${size}" viewBox="-25 -25 50 50">${icon[pay.current.ww].svg}</svg>`;

                                pay.Txt.MosmixIcon = MosmixIcon.icon;
                                {
                                    let ww = pay.current.ww;
                                    const icon = MosmixIcon.icon;
                                    if ((ww >= 0) && icon[ww]) {
                                        pay.Txt.iconWw = icon[ww].svg;
                                    }
                                }
                                this.#outData(msg);
                                this.dashboard(pay, msg.NodeStation);
                                // this.prom = null;
                                // console.log("then");
                            } catch (err) {
                                this.warn(err.message);
                            }
                        }
                    });
                })
                .catch((message) => {
                    this.warn(RED._("redhm-weather.warn.requestFailed", { 'error': message }));
                })
                .finally(() => {
                    this.prom.done = true;
                    // console.log("fin");
                    if (done)
                        done();
                })
        };

        dashboard(pay, NodeStation) {
            let cur = pay.current;
            //debugger;

            {
                let subIcon = "";
                const mx = Mosmix.GetWW_Icon(cur.ww);
                if (mx) {
                    subIcon = `<i class="ww" style="padding: 10%;"><svg xmlns="http://www.w3.org/2000/svg" version="1.1" width="90%" height="90%" viewBox="-25 -25 50 50" transform="translate(2,2)">${mx.svg}</svg></i>`
                }
                let moon = ""
                if (cur.Moon && cur.Moon.wi) {
                    moon = cur.Moon.wi;
                }
                let sunOut = "";
                const sun = Sun_cl.makeSunTxt(cur, pay.timeZone, pay.timezone_offset);
                if (sun) {
                    if (sun.sunUpDown)
                        sunOut = sun.sunUpDown;
                }

                const fc = makeForecast(pay, "DWD");
                this.#outForecast({
                    template: fc.all,
                    kurz: fc.akt,
                    NodeStation: NodeStation,
                    Icon: Rwi.GetSvg_Wetter(0, 0, 90, 90, wDataToStufe(cur, pay)),
                    subIcon: subIcon,
                    subTemp: Rwi.GetSvg_Thermo(4, 4, 40, 40, 25),
                    subMond: Rwi.GetSvg_Mond(4, 4, 40, 40, cur.Moon),
                    subSonne: Rwi.GetSvg_Sonne(4, 4, 40, 40),
                    subWind: Rwi.GetSvg_Wind(4, 4, 40, 40, cur.wind_deg, Math.max(cur.wind_speed, cur.wind_gust)),
                    sun: sunOut,
                    moon: moon
                });
            }
        }

        #outData(msg) {
            this.send(msg);
        }

        #outForecast(msg) {
            this.send([null, msg]);
        }

        outSpezial(msg) {
            this.send([null, null, msg]);
        }

        #close(done) {
            // console.log("close: " + this.mStation);
            /*            if (this.intervalId !== null) {
                            clearInterval(this.intervalId);
                        }*/
            if (done)
                done();
        };

        /*        if (!this.omitMessageOnStart) {
                    this.emit("input", {
                        payload: {}
                    });
                }
                */
    }
    RED.nodes.registerType("redhm-weather-dwd", RedHM_DWD, { credentials: { mosmixStation: { type: "text" } } });

    class DWD_Wetterwarnung {
        constructor(nodeConfig) {
            RED.nodes.createNode(this, nodeConfig);

            this.Cf = {};
            this.Cf.nodeId = this.id;
            this.Cf.nodeName = nodeConfig.name || "DWD_Wetterwarnung";

            this.Cf.WarnCellId = Number(this.credentials.WarnCellId);

            const tTime = nodeConfig.Intervall || "5/30 * * * *";
            if (tTime != "Aus")
                this.task_Intervall = scheduleTask(tTime, (timestamp) => {
                    // console.log(`Task triggered at ${timestamp}`);
                    this.#input({});
                })

            this.on('input', this.#input);
            this.on('close', this.#close);
            this.status({ fill: "yellow", shape: "ring", text: "ready" });
        }

        #input(mg, send, done) {
            https.get('https://www.dwd.de/DWD/warnungen/warnapp/json/warnings.json', (res) => {
                // console.log('statusCode:', res.statusCode);
                if (res.statusCode != 200) {
                    Logg.error(this, "statusCode: " + res.statusCode);
                    return;
                }
                //console.log('headers:', res.headers);
                let data = "";

                res.on('data', (chunk) => {
                    // console.log(chunk);
                    data += chunk;
                });
                res.on('end', () => {
                    try {
                        //let json = JSON.parse(data);
                        //console.log(json);

                        let json = JSON.parse(data.replace(/^warnWetter\.loadWarnings\(/, '').replace(/\);\s*$/, ''));
                        // Console.log(json);

                        let region = {};

                        for (const id of Object.keys(json.warnings)) {
                            //debugger;
                            json.warnings[id].forEach((warn, i) => {
                                region[id] = { state: warn.state, region: warn.regionName };
                            });
                        }

                        let warnings = json.warnings[this.Cf.WarnCellId] || [];
                        // msg.vorabInformation = json.vorabInformation[this.Cf.WarnCellId] || [];

                        const dwd = {
                            COLORS: {
                                0: '#c5e566',
                                1: '#b39ddb',
                                2: '#ffeb3b',
                                3: '#fb8c00',
                                4: '#e53935',
                                5: '#880e4f',
                                10: '#cc99ff',
                                20: '#fe68fe',
                                30: '#c5e566',
                                50: '#cc99ff',
                            },
                            LEVEL_ORDER: {
                                0: 0,
                                1: 1,
                                2: 2,
                                3: 3,
                                4: 4,
                                5: 5,
                                10: 1.5,
                                20: 1.4,
                                50: 1.5,
                            },
                            WARN_TYPES: {
                                0: 'gewitter',
                                1: 'sturm',
                                2: 'regen',
                                3: 'schnee',
                                4: 'nebel',
                                5: 'frost',
                                6: 'glatteis',
                                7: 'tauwetter',
                                8: 'hitze',
                                9: 'uv',
                                10: 'kueste',
                                11: 'binnensee'
                            },
                            WARN_COLORS_CLASS: {
                                0: 'green',
                                1: 'violet',
                                2: 'yellow',
                                3: 'orange',
                                4: 'red',
                                5: 'darkred',
                                10: 'violet_hitze',
                                20: 'violet_uv',
                                50: 'violet_hitze',
                            },
                            WARN_COLORS: {
                                0: 'gelb',
                                1: 'gelb',
                                2: 'gelb',
                                3: 'orange',
                                4: 'rot',
                                5: 'dunkelrot',
                                10: 'violet_hitze',
                                20: 'violet_uv',
                                50: 'violet_hitze',
                            }
                        };

                        const now = (new Date()).getTime();
                        // Logg.warn ( this,new Date().toLocaleString(Language, { dateStyle: 'medium', timeStyle: "short" }));

                        let pay = "";
                        let payTime = "";
                        let payColor = "grey";
                        let iconPay = "";
                        let maxLevel = 0;
                        const events = [];
                        let html = `<style>
.dwd-icon-container {display: inline-block;}
.dwd-event {
    display: inline-block;
    vertical-align: 30px;
    font-weight: bold;
    padding-left: 20px;
}
.dwd-icon,
.dwd-icon-frame {position: absolute;}
.dwd-icon-container,
.dwd-icon,
.dwd-icon-frame {width: 50px;height: 50px;}
.dwd-time {color:grey;}
.dwd-timeAkt {color:LightSalmon;}
</style>
`;

                        warnings.forEach((warn, i) => {
                            /*    send([null, {
                                    payload: warn
                                }]);*/
                            warn.pre = (warn.start && (now < warn.start));
                            warn.past = (warn.end && (now > warn.end));
                            warn.active = !(warn.pre || warn.past);
                            warn.typeName = dwd.WARN_TYPES[warn.type];
                            warn.color = dwd.COLORS[warn.level];
                            warn.order = dwd.LEVEL_ORDER[warn.level];
                            warn.icon = 'http://www.dwd.de/DWD/warnungen/warnapp/viewer/img/warndreieck/warn_icons_' + warn.typeName + '.png';
                            let col = dwd.WARN_COLORS[warn.level];
                            if (col == 'undefined') {
                                Logg.warn(this, "DWD");
                                col = "gelb";
                            }
                            warn.frame = 'http://www.dwd.de/DWD/warnungen/warnapp/viewer/img/warndreieck/' + col + '.png';
                            const start = new Date(warn.start);
                            warn.time = start.toLocaleString(Language, { dateStyle: 'medium', timeStyle: "short" });
                            if (warn.end) {
                                const end = new Date(warn.end);
                                //if (start.getDay() === end.getDay()) {
                                warn.time += ' - ' + end.toLocaleString(Language, { timeStyle: "short" });
                                //} else {
                                //warn.time += ' - ' + end.toLocaleTimeString(Language, { dateStyle: 'medium', timeStyle: "short" });
                                //}
                            }
                            if (!warn.past) {
                                let timeStyle = "dwd-time";
                                if (warn.active) {
                                    timeStyle = "dwd-timeAkt";
                                    if (warn.order > maxLevel) {
                                        maxLevel = warn.order;
                                        pay = warn.event;
                                        iconPay = RwiX.GetSvg_WetterWarnung(-20, -5, 60, 60, warn.typeName, warn.color);
                                        payColor = warn.color;
                                        payTime = warn.time;
                                    }
                                }
                                if (!events.includes(warn.event))
                                    events.push(warn.event);

                                {
                                    const icon = RwiX.GetSvg_WetterWarnung(-20, -5, 60, 60, warn.typeName, warn.color);
                                    html += `
<div class="dwd-warning-container">
    <div class="dwd-icon-container">${icon}</div>
    <div class="dwd-event">${warn.event}</div>
    <div class="${timeStyle}">${warn.time}</div>
    <div class="dwd-description">${warn.description}</div>
</div>
`;
                                }
                            }

                        });

                        this.send({
                            payload: pay,
                            icon: iconPay,
                            color: payColor,
                            payTime: payTime,
                            region: region,
                            count: warnings.length,
                            events: events.join(', '),
                            warnings,
                            //html: html.join('\n'),
                            template: html
                        });
                        this.status({ fill: "green", shape: "dot", text: pay });
                    } catch (e) {
                        Logg.error(this, 'Unexpected response: ' /*+ data*/);
                    }
                });

            }).on('error', (e) => {
                Logg.error(this, e);
            });
            if (done)
                done();
        }

        #close(done) {
            if (this.task_Intervall)
                this.task_Intervall.stop();
            this.status({ fill: "yellow", shape: "dot", text: "stop" });
            done();
        }

    }
    RED.nodes.registerType('dwd-wetterwarnung', DWD_Wetterwarnung, { credentials: { WarnCellId: { type: "text" } } });

    class RedHM_PosConfig {
        constructor(nodeConfig) {
            RED.nodes.createNode(this, nodeConfig);
            this.PosConfig = {
                Name: nodeConfig.name,
                lat: Number(this.credentials.lat), lon: Number(this.credentials.lon), alt: Number(nodeConfig.alt),
                timeZone: nodeConfig.timeZone
            };
            // Logg.debug(this, "RedHM_PosSetup");

            this.on('close', this.#close);
        }

        #close(done) {
            done();
        };
    }

    RED.nodes.registerType('redhm-pos_config', RedHM_PosConfig, { credentials: { lat: { type: "text" }, lon: { type: "text" } } });

    class RedHM_PosGet {
        constructor(nodeConfig) {
            RED.nodes.createNode(this, nodeConfig);
            this.pos_config = RED.nodes.getNode(nodeConfig.pos_config);
            if (!this.pos_config) {
                this.error(RED._('missing setup config'));
                return;
            }
            this.topic = nodeConfig.topic;
            this.PosConfig = this.pos_config.PosConfig;
            this.isTag = undefined;
            // Logg.debug(this, "RedHM_PosSetup");

            this.on('input', this.#input);
            this.on('close', this.#close);
            // this.status({ fill: "yellow", shape: "ring", text: "ready" });
        }

        #input(mg, send, done) {
            if (this.topic)
                mg.topic = this.topic;

            let date = new Date();
            if (mg.time)
                date = new Date(mg.time);

            const sr = new Sun_cl(date.getTime() / UNIX_MUL, this.PosConfig.lat, this.PosConfig.lon);
            let info = { ...this.PosConfig, ... { sunrise: sr.sunrise, sunset: sr.sunset, IsTag: sr.IsTag, change: (sr.IsTag != this.IsTag) } };
            this.IsTag = sr.IsTag;
            let aa = Sun_cl.makeSunTxt(sr, this.PosConfig.timeZone);
            let col = "yellow";
            if (!sr.IsTag)
                col = "black";
            this.status({ fill: col, shape: "dot", text: aa.sunrise + " - " + aa.sunset })
            let immer = RED.util.cloneMessage(mg);
            immer.payload = info;
            send(immer);
            mg.PosInfo = info;
            if (sr.IsTag)
                send([null, mg]);
            if (sr.IsTag == false)
                send([null, null, mg]);
            done();
        }

        #close(done) {
            done();
        }
    }
    RED.nodes.registerType('redhm-pos_get', RedHM_PosGet);

    function fixInMsg(node, typ, msg) {
        let Bez = node.Bez;
        if (node.topic)
            msg.topic = node.topic;
        let lat = node.PosConfig.lat;
        let lon = node.PosConfig.lon;
        msg.WeatherTyp = typ;
        msg.NodeStation = true;
        if (typeof msg.payload == 'object')
            if (typeof msg.payload.lat == 'number')
                if (typeof msg.payload.lon == 'number') {
                    lon = msg.payload.lon;
                    lat = msg.payload.lat;
                    if (msg.payload.Name)
                        Bez = msg.WeatherTyp + " " + msg.payload.Name;
                    else
                        Bez = msg.WeatherTyp + " " + lon.toFixed(2) + "/" + lat.toFixed(2);
                    if ((lat != node.PosConfig.lat) || (lat != node.PosConfig.lat))
                        msg.NodeStation = false;
                }

        if (!msg.payload)
            msg.payload = {};
        msg.call = { Bez: Bez, lat: lat, lon: lon };
    }

    var Ow_LastMSg = {};

    class RedHM_OW {
        constructor(nodeConfig) {
            RED.nodes.createNode(this, nodeConfig);
            this.pos_config = RED.nodes.getNode(nodeConfig.pos_config);
            if (!this.pos_config) {
                this.error(RED._('missing setup config'));
                return;
            }
            this.topic = nodeConfig.topic;
            this.Bez = nodeConfig.name;
            if (!this.Bez)
                this.Bez = nodeConfig.defaultName;
            this.PosConfig = this.pos_config.PosConfig;
            // Logg.debug(this, "RedHM_PosSetup");
            this.apiKey = this.credentials.apiKey;
            this.UnitTemp = nodeConfig.UnitTemp;
            this.UnitSpeed = nodeConfig.UnitSpeed;

            this.on('input', this.#input);
            this.on('close', this.#close);
            // this.status({ fill: "yellow", shape: "ring", text: "ready" });
        }

        #input(msg, send, done) {
            fixInMsg(this, "OW", msg);
            let api = {
                "lat": msg.call.lat,
                "lon": msg.call.lon,
                "appid": this.apiKey,

                "units": "metric",
                "lang": Language,
                "exclude": "minutely"
            }

            let url = "https://api.openweathermap.org/data/2.5/onecall";
            // let url = "https://api.openweathermap.org/data/3.0/onecall";
            {
                let addUrl = "";
                for (const key of Object.keys(api)) {
                    if (addUrl)
                        addUrl += "&";
                    addUrl += key + "=" + api[key];
                }
                if (addUrl)
                    url += "?" + addUrl;
            }
            const updateMin = 60 * 5;
            let newId = msg.call.lon + "/" + msg.call.lat;
            for (const id of Object.keys(Ow_LastMSg)) {
                let obj = Ow_LastMSg[id];
                let diff = new Date().getTime() / UNIX_MUL - obj.payload.current.dt;
                if (diff >= updateMin) {
                    delete Ow_LastMSg[id];
                    continue;
                }
                // if (obj.payload && obj.payload.current && obj.payload.current.dt) {
                if (newId == id) {
                    // console.log("Last: " + diff.toFixed(1));
                    this.#parseData(RED.util.cloneMessage(obj), send, done);
                    return;
                }
            }
            httpGet(url, "OW")
                .then((res) => {
                    // console.log("get");
                    msg.payload = JSON.parse(res.data);
                    msg.payload.Bez = msg.call.Bez;
                    msg.payload.update = { Akt: msg.payload.current.dt, Next: msg.payload.current.dt + updateMin };
                    Ow_LastMSg[newId] = RED.util.cloneMessage(msg);
                    this.#parseData(msg, send, done);
                })
                .catch((error) => {
                    this.#ErrOut(error);
                    done();
                });
        }

        #parseData(msg, send, done) {
            let pay = msg.payload;
            let cur = pay.current;
            {
                if (typeof cur.temp != "number")
                    return;

                // console.log(Bez);
                pay.timeZone = pay.timezone;
                delete pay.timezone;

                if (cur.wind_deg == 360)
                    cur.wind_deg = 0;

                cur.IsTag = Sun_cl.getIsTag(cur.dt, cur.sunrise, cur.sunset);

                pay.Dps = getDpsInit();
                {
                    let arrKey = [];
                    function pObj(ele) {
                        for (const key of (Object.keys(ele))) {
                            let obj = ele[key];
                            switch (key) {
                                case 'rain':
                                case 'snow':
                                    if (obj) {
                                        if (obj["1h"])
                                            ele[key] = obj["1h"];
                                    }
                                    if (!ele[key])
                                        ele[key] = 0;
                                    break;
                                default:
                                    break;
                            }
                            arrKey.push(key);
                        }
                    }
                    pObj(cur);
                    pay.hourly.forEach(function (ele) {
                        pObj(ele);
                    });
                    pay.daily.forEach(function (ele) {
                        pObj(ele);
                    });

                    arrKey.forEach(function (key) {
                        let unit = "";
                        let desc = "";
                        for (const ele of Object.keys(MosmixElement)) {
                            let me = MosmixElement[ele];
                            if (me && me.IdOut == key) {
                                unit = me.Unit;
                                desc = Mosmix.getDesc(ele);
                            }
                        }
                        switch (key) {
                            case 'wind_speed':
                            case 'wind_gust':
                                unit = "m/s";
                                break;

                            case 'temp':
                            case 'dew_point':
                            case 'feels_like':
                                unit = "C";
                                break;

                            case 'precipitation':
                                unit = "mm";
                                break;

                            case 'pressure':
                                unit = "hPa";
                                break;

                            default:
                                break;
                        }
                        if (!pay.Dps[key])
                            pay.Dps[key] = { unit: unit, desc: desc };
                    });

                    delete pay.Dps.moonrise;
                    delete pay.Dps.moonset;
                    delete pay.Dps.moon_phase;
                    delete pay.Dps.dt;
                    delete pay.Dps.sunrise;
                    delete pay.Dps.sunset;
                    delete pay.Dps.weather;
                    delete pay.Dps.pop;
                }

                {
                    function rrpObj2(ele) {
                        for (const key of (Object.keys(ele))) {
                            let wert = ele[key];
                            if (pay.Dps[key]) {
                                let dp = pay.Dps[key];
                                if (dp.unit) {
                                    if (typeof wert == 'object') {
                                        for (const wKey of (Object.keys(wert))) {
                                            let wert2 = wert[wKey];
                                            let neu = fixUnit2(this, dp.unit, wert2);
                                            ele[key][wKey] = neu;
                                            // if (neu != wert2)
                                            // console.log(key + " " + dp.unit + " " + wert2 + " " + neu);
                                        }
                                    }
                                    else {
                                        let neu = fixUnit2(this, dp.unit, wert);
                                        ele[key] = neu;
                                        // if (neu != wert)
                                        // console.log(key + " " + dp.unit + " " + wert + " " + neu);
                                    }
                                }
                            }
                        }
                        if (!ele.rain) ele.rain = 0;
                        if (!ele.snow) ele.snow = 0;
                        ele.precipitation = ele.rain + ele.snow;
                    }
                    let pObj2 = rrpObj2.bind(this);
                    pObj2(cur);
                    pay.hourly.forEach(function (ele) {
                        pObj2(ele);
                    });
                    pay.daily.forEach(function (ele) {
                        pObj2(ele);
                    });
                }
                for (const key of (Object.keys(pay.Dps))) {
                    pay.Dps[key].unit = UnitGetOut(this, pay.Dps[key].unit);
                }
                if (!cur.wind_gust) cur.wind_gust = pay.hourly[0].wind_gust;
                if (!cur.wind_gust) cur.wind_gust = 0;

                if (!cur.rain) pay.hourly[0].rain;
                if (!cur.snow) pay.hourly[0].snow;

                // this.outSpezial({ dps: pay.Dps, MosmixElement: MosmixElement });
                cur.description = cur.weather[0].description;
                pay = addData(pay);
            }

            // dashboard
            {
                msg.payload = pay;

                this.#outData(msg);
                const fc = makeForecast(pay, "OW");
                this.#outForecast({ template: fc.all, kurz: fc.akt, NodeStation: msg.NodeStation });
            }

            this.status({ fill: "green", shape: "dot", text: new Date().toLocaleString(Language) });
            done();
        }

        #outData(msg) {
            this.send(msg);
        }

        #outForecast(msg) {
            this.send([null, msg]);
        }

        #ErrOut(txt = "") {
            this.error(txt);
            this.status({ fill: "red", shape: "dot", text: "Fehler: " + txt });
        }

        #close(done) {
            done();
        }
    }
    RED.nodes.registerType('redhm-weather-ow', RedHM_OW, { credentials: { apiKey: { type: "text" } } });

    var Meteo_LastMSg = {};

    class RedHM_Meteo {
        constructor(nodeConfig) {
            RED.nodes.createNode(this, nodeConfig);
            this.pos_config = RED.nodes.getNode(nodeConfig.pos_config);
            if (!this.pos_config) {
                this.error(RED._('missing setup config'));
                return;
            }
            this.topic = nodeConfig.topic;
            this.Bez = nodeConfig.name;
            if (!this.Bez)
                this.Bez = nodeConfig.defaultName;
            this.PosConfig = this.pos_config.PosConfig;
            // Logg.debug(this, "RedHM_PosSetup");
            this.model = nodeConfig.model;
            if (!this.model)
                this.model = "forecast";

            this.UnitTemp = nodeConfig.UnitTemp;
            this.UnitSpeed = nodeConfig.UnitSpeed;

            this.on('input', this.#input);
            this.on('close', this.#close);
            // this.status({ fill: "yellow", shape: "ring", text: "ready" });
        }

        #input(msg, send, done) {
            fixInMsg(this, "Meteo " + this.model, msg);
            let newId = this.model + "/" + msg.call.lon + "/" + msg.call.lat;
            for (const id of Object.keys(Meteo_LastMSg)) {
                let obj = Meteo_LastMSg[id];
                let diff = new Date().getTime() / UNIX_MUL - obj.Meteo.current.time;
                if (diff >= obj.Meteo.current.interval) {
                    delete Meteo_LastMSg[id];
                    continue;
                }
                // if (obj.payload && obj.payload.current && obj.payload.current.dt) {
                if (newId == id) {
                    // console.log("Last: " + diff.toFixed(1));
                    this.#parseData(RED.util.cloneMessage(obj), send, done);
                    return;
                }
            }
            {
                // ,apparent_temperature,is_day, surface_pressure
                const get = "temperature_2m,relative_humidity_2m,dew_point_2m,precipitation_probability,precipitation,rain,showers,snowfall,weather_code,cloud_cover,pressure_msl,visibility,wind_speed_10m,wind_direction_10m,wind_gusts_10m,uv_index";
                const getD = "weather_code,temperature_2m_max,temperature_2m_min,uv_index_max,precipitation_sum,precipitation_probability_max,rain_sum,showers_sum,snowfall_sum,wind_speed_10m_max,wind_gusts_10m_max,wind_direction_10m_dominant,cloud_cover_mean,relative_humidity_2m_mean,pressure_msl_mean,visibility_mean,dew_point_2m_mean";
                let urlMeteo = `https://api.open-meteo.com/v1/${this.model}?latitude=${msg.call.lat}&longitude=${msg.call.lon}&current=${get}&hourly=${get}&forecast_hours=48&daily=${getD}&timeformat=unixtime&timezone=auto&forecast_days=8`;
                httpGet(urlMeteo, "Meteo")
                    .then((res) => {
                        // console.log("get");
                        msg.Meteo = JSON.parse(res.data);
                        msg.payload = {
                            "lat": msg.call.lat,
                            "lon": msg.call.lon,
                            timeZone: msg.Meteo.timezone
                        };
                        msg.payload.Bez = msg.call.Bez;
                        msg.payload.update = { Akt: msg.Meteo.current.time, Next: msg.Meteo.current.time + msg.Meteo.current.interval };
                        Meteo_LastMSg[newId] = RED.util.cloneMessage(msg);
                        this.#parseData(msg, send, done);
                    })
                    .catch((error) => {
                        this.#ErrOut(error);
                        done();
                    });

            }
        }

        #parseData(msg, send, done) {
            let Meteo = msg.Meteo;
            let pay = msg.payload;
            let that = this;

            function fixId(id) {
                const newId = {
                    time: "dt", temperature_2m: "temp", relative_humidity_2m: "humidity", dew_point_2m: "dew_point", weather_code: "ww", cloud_cover: "clouds", pressure_msl: "pressure",
                    wind_speed_10m: "wind_speed", wind_direction_10m: "wind_deg", wind_gusts_10m: "wind_gust", uv_index: "uvi",
                    wind_speed_10m_max: "wind_speed", wind_direction_10m_dominant: "wind_deg", wind_gusts_10m_max: "wind_gust", uv_index_max: "uvi",
                    rain_sum: "rain", snowfall: "snow", snowfall_sum: "snow", showers_sum: "showers", precipitation_sum: "precipitation", precipitation_probability: "precipitation_perc", precipitation_probability_max: "precipitation_perc", cloud_cover_mean: "clouds", cloud_cover_min: "clouds",
                    relative_humidity_2m_mean: "humidity", dew_point_2m_mean: "dew_point", pressure_msl_mean: "pressure", visibility_mean: "visibility"
                };
                if (newId[id])
                    return newId[id];
                return id;
            }
            function fixObj(obj) {
                // cm -> mm
                obj = { ...obj, ... new Sun_cl(obj.dt, pay.lat, pay.lon).get() };
                obj.description = Mosmix.GetWW_Desc(obj.ww);
                obj.snow = Math.round(obj.snow * 10 / 7 * 100) / 100;
                // if (obj.precipitation != obj.rain + obj.showers + obj.snow) // Teilweise falsch
                // console.log(obj.precipitation, obj.rain, obj.showers, obj.snow);
                return obj;
            }
            pay.current = {};
            pay.Dps = getDpsInit();
            for (let id of Object.keys(Meteo.current)) {
                function addDp(key) {
                    let unit = "";
                    let desc = "";
                    for (const ele of Object.keys(MosmixElement)) {
                        let me = MosmixElement[ele];
                        if (me && me.IdOut == key) {
                            unit = me.Unit;
                            desc = Mosmix.getDesc(ele);
                        }
                    }
                    switch (key) {
                        case 'wind_speed':
                        case 'wind_gust':
                            unit = "km/h";
                            break;

                        case 'temp':
                        case 'dew_point':
                        case 'feels_like':
                            unit = "C";
                            break;
                        case 'pressure':
                            unit = "hPa";
                            break;

                        default:
                            break;
                    }
                    if (!pay.Dps[key]) {
                        pay.Dps[key] = { unit: unit, desc: desc };
                        pay.Dps[key].unit = UnitGetOut(that, pay.Dps[key].unit);
                    }
                }
                const obj = Meteo.current[id];
                const idFix = fixId(id);
                addDp(idFix);
                pay.current[idFix] = obj;
            }
            pay.current = fixObj(pay.current);

            function copyArrToPay(name) {
                pay[name] = [];
                for (let id of Object.keys(Meteo[name])) {
                    const obj = Meteo[name][id];
                    for (let x = 0; x < obj.length; x++) {
                        if (obj[x] == null)
                            continue;

                        if (!pay[name][x])
                            pay[name][x] = {};
                        const useID = fixId(id);
                        switch (useID) {
                            case "temperature_2m_max":
                                if (!pay[name][x]["temp"])
                                    pay[name][x]["temp"] = {};
                                pay[name][x]["temp"].max = obj[x];
                                break;
                            case "temperature_2m_min":
                                if (!pay[name][x]["temp"])
                                    pay[name][x]["temp"] = {};
                                pay[name][x]["temp"].min = obj[x];
                                break;
                            default:
                                pay[name][x][useID] = obj[x];
                                break;
                        }
                    }
                }
                for (let x = 0; x < pay[name].length; x++) {
                    if (Object.keys(pay[name][x]).length <= 1) {
                        delete pay[name][x];
                        continue;
                    }
                    pay[name][x] = fixObj(pay[name][x]);
                    if (name == "daily")
                        pay[name][x].IsTag = undefined;
                }
            }
            copyArrToPay("hourly");
            copyArrToPay("daily");
            addData(pay)
            // msg.Meteo = Meteo;
            delete msg.Meteo;
            // dashboard
            {
                // msg.payload = pay;

                this.#outData(msg);
                const fc = makeForecast(pay, "Meteo");
                this.#outForecast({ template: fc.all, kurz: fc.akt, NodeStation: msg.NodeStation });
            }

            this.status({ fill: "green", shape: "dot", text: new Date().toLocaleString(Language) });
            done();
        }

        #outData(msg) {
            this.send(msg);
        }

        #outForecast(msg) {
            this.send([null, msg]);
        }

        #ErrOut(txt = "") {
            this.error(txt);
            this.status({ fill: "red", shape: "dot", text: "Fehler: " + txt });
        }

        #close(done) {
            done();
        }
    }
    RED.nodes.registerType('redhm-weather-meteo', RedHM_Meteo);
}

class Sun_cl {
    constructor(date, lat, lon) {
        let useDate;
        if (!date)
            useDate = new Date();
        else
            useDate = new Date(date * UNIX_MUL);

        const nLat = Number(lat);
        const nLon = Number(lon);
        function i(rs) {
            const cE = 90.8333;
            function n(t) { return Math.sin(2 * t * Math.PI / 360) }
            function a(t) { return 360 * Math.acos(t) / (2 * Math.PI) }
            function r(t) { return Math.cos(2 * t * Math.PI / 360) }
            function u(t, e) { var n = t % e; return n < 0 ? n + e : n }
            const fu = Math.ceil((useDate.getTime() - new Date(useDate.getFullYear(), 0, 1).getTime()) / 864e5);
            const s = nLon / 15;
            const l = rs ? fu + (6 - s) / 24 : fu + (18 - s) / 24;
            const g = .9856 * l - 3.289;
            const v = u(g + 1.916 * n(g) + .02 * n(2 * g) + 282.634, 360);
            const P = .91764 * (v, Math.tan(2 * v * Math.PI / 360));
            let c = u(360 / (2 * Math.PI) * Math.atan(P), 360);
            c += 90 * Math.floor(v / 90) - 90 * Math.floor(c / 90);
            c /= 15;
            const I = .39782 * n(v);
            const S = r((I, 360 * Math.asin(I) / (2 * Math.PI)));
            const d = (r(cE) - I * n(nLat)) / (S * r(nLat));
            const w = u((rs ? 360 - a(d) : a(d)) / 15 + c - .06571 * l - 6.622 - nLon / 15, 24);
            const T = Date.UTC(useDate.getFullYear(), useDate.getMonth(), useDate.getDate());
            return new Date(T + 36e5 * w)
        }
        this.sunrise = Math.round(i(!0).getTime() / UNIX_MUL);
        this.sunset = Math.round(i(!1).getTime() / UNIX_MUL);
        this.IsTag = Sun_cl.getIsTag(date, this.sunrise, this.sunset);
    }

    get() {
        return this;
    }

    static getIsTag(dt, sunrise, sunset) {
        if (!sunrise || !sunset || Number.isNaN(sunset) || Number.isNaN(sunrise))
            return undefined;
        if (dt >= sunrise)
            if (dt <= sunset)
                return true;
        return false;
    }
    static addsunUpDown(Txt, incl = Rwi) {
        if (!Txt.sunrise)
            return "";
        return '<span style="vertical-align:top">' + incl.GetSvg_Sonne(0, 0, C.INFOTXTSIZE, C.INFOTXTSIZE) + `<span> &#8673;&nbsp;${Txt.sunrise}&nbsp;&#8675;&nbsp;${Txt.sunset}&nbsp;>&nbsp;${Txt.sun}</span></span>`;
    }

    static makeSunTxt(cur, timeZone, timeOff = 0, incl = Rwi) {
        let Txt = {};
        if (!timeZone)
            return Txt;
        if (!cur.sunrise)
            return Txt;
        if (!cur.sunset)
            return Txt;

        // let xx = Intl.DateTimeFormat("ia", { timeZoneName: "short", timeZone, }).formatToParts(); // "ia" =  Interlingua

        Txt.sunrise = timeShort(cur.sunrise, timeZone);
        Txt.sunset = timeShort(cur.sunset, timeZone);
        if (false)
            if (timeOff) {
                let dateObject = new Date((cur.sunrise + timeOff) * UNIX_MUL);
                console.log(dateObject.toLocaleString("en", { hour12: C.Hour12, hour: "numeric", minute: "2-digit" }).toLowerCase() + " " + Txt.sunrise);
            }

        if (Number.isNaN(cur.sunset) || Number.isNaN(cur.sunrise)) {
            Txt.sun = "";
        }
        else {
            let ss = cur.sunset;
            if (ss < cur.sunrise)
                ss += 24 * 60 * 60;
            let diff = (ss - cur.sunrise);
            if (diff < 0)
                console.log("hmmm");
            let stunde = Math.floor(diff / (60 * 60));
            let minute = Math.round((diff - (stunde * 60 * 60)) / 60);
            let add = "";
            if (minute < 10)
                add = "0";
            Txt.sun = stunde + ":" + add + minute;
            // console.log ( `${Txt.sunrise} ${Txt.sunset} ${Txt.sun}` );
            // <i class="fa fa-arrow-up"></i> <i class="fa fa-arrow-down"></i> <i class="wi wi-owm-01d"></i>
            Txt.sunUpDown = this.addsunUpDown(Txt, incl);
        }
        return Txt;
    }
}

function httpGet(url, typ = "") {
    if (!url)
        return null;

    return new Promise((resolve, reject) => {
        const httpUse = (url.indexOf("https://") === 0) ? https : http;

        // debugger;

        httpUse.get(url, (res) => {
            // console.log("statusCode:", res.statusCode);
            if (res.statusCode != 200) {
                reject("statusCode: " + res.statusCode);
                return;
            }
            // console.log("headers:", res.headers);
            let data = "";

            res.on("data", (chunk) => {
                // console.log(chunk);
                data += chunk;
            });
            res.on("end", () => {
                // console.log(data);
                resolve({ typ: typ, data: data });
            });

        }).on("error", (e) => {
            reject(e);
        });
    });
}

function timeZoneNameShort(timeZone) {
    let xx = Intl.DateTimeFormat(Language, { timeZoneName: "short", timeZone: timeZone, }).formatToParts();
    let UTC = xx.find((i) => i.type === "timeZoneName").value;
    if (!UTC)
        console.log("Err:", timeZone)
    return UTC;
}

function timeShort(UNIX_timestamp, timeZone, onlyHour = false) {
    if (Number.isNaN(UNIX_timestamp))
        return "";
    let dateObject = new Date(UNIX_timestamp * UNIX_MUL);
    if (onlyHour)
        return dateObject.toLocaleString("en", { timeZone: timeZone, hour12: C.Hour12, hour: "numeric" }).toLowerCase() + ":00";
    else
        return dateObject.toLocaleString(Language, { timeZone: timeZone, hour12: C.Hour12, hour: "numeric", minute: "2-digit" }).toLowerCase();
}

function dayName(unixTime, timeZone) {
    let dateObject = new Date(unixTime * UNIX_MUL);
    return dateObject.toLocaleString(Language, { /*timeZone: timeZone,*/ weekday: "short" });
}

function TagImJahr(date) {
    return (Date.UTC(date.getFullYear(), date.getMonth(), date.getDate()) - Date.UTC(date.getFullYear(), 0, 0)) / 24 / 60 / 60 / 1000;
}

function addData(pay) {
    if (!pay.Txt)
        pay.Txt = {};

    let Txt = pay.Txt;
    pay.current = { ...pay.current, ...Moon_Cl.get(pay.lat, new Date(pay.current.dt * UNIX_MUL)) };

    let cur = pay.current;
    if (!pay.timeZone)
        pay.timeZone = Intl.DateTimeFormat().resolvedOptions().timeZone;

    Txt.temp = cur.temp.toFixed(1) + " " + pay.Dps.temp.unit;

    cur.wind_cardinal = degreesToCardinal(cur.wind_deg);
    Txt.wind_speed = cur.wind_speed.toFixed();
    if (cur.wind_gust)
        Txt.wind_speed += "/" + cur.wind_gust.toFixed();
    Txt.wind_speed += " " + pay.Dps.wind_speed.unit;
    // Txt.wind_speed += " " + cur.wind_cardinal;

    function make(ww) {
        return Math.round(ww.min) + " < " + Math.round(ww.schnitt) + " > " + Math.round(ww.max) + " " + pay.Dps.temp.unit;
    }
    function schnitt(hourly, anz) {
        if (anz < hourly.lenght)
            return;

        let sch = {
            max: Number.NEGATIVE_INFINITY,
            min: Number.POSITIVE_INFINITY
        };

        // node.send({ payload: [] });
        let sum = 0;

        for (let index = 0; index < anz; index++) {
            const ele = hourly[index];
            if (ele.temp < sch.min)
                sch.min = ele.temp;
            if (ele.temp > sch.max)
                sch.max = ele.temp;
            sum += ele.temp;
        }
        sch.schnitt = Math.round(sum / anz * 10) / 10;
        sch.max = Math.round(sch.max * 10) / 10;
        sch.min = Math.round(sch.min * 10) / 10;
        return sch;
    }

    pay.Temp24h = schnitt(pay.hourly, 24);
    Txt.Temp24h = make(pay.Temp24h);
    pay.Temp48h = schnitt(pay.hourly, 48);
    Txt.Temp48h = make(pay.Temp48h);

    pay.Txt = { ...Txt, ...Sun_cl.makeSunTxt(cur, pay.timeZone, pay.timezone_offset) };
    return pay;
}

function makeForecast(pay, from) {
    // let incl = RwiX.inCl;
    let incl = Rwi;

    function info2(element) {
        let rain = 0;
        if (typeof element.rain == 'object')
            rain += element.rain.sum;
        else
            rain += element.rain

        if (element.showers) {
            if (typeof element.showers == 'object')
                rain += element.showers.sum;
            else
                rain += element.showers
        }

        rain = element.precipitation;
        let snow = element.snow;
        if (typeof snow == 'object')
            snow = snow.sum;

        let wind_gust = element.wind_gust;
        if (typeof wind_gust == 'object')
            wind_gust = wind_gust.max;

        let txt = "";
        let icon = "";
        if (element.precipitation >= 1) {
            txt = Math.round(element.precipitation);
            icon = '<i class="fa fa-tint"></i>';
            icon = incl.GetSvg_Tropfen(0, -2, C.INFOTXTSIZE, C.INFOTXTSIZE);
            //            icon += incl.GetSvg_Schneeflocke(0, 4, C.INFOTXTSIZE, C.INFOTXTSIZE) ;
        }
        if (element.snow >= 1) {
            // txt = Math.round(snow);
            //            icon = '<i class="fa fa-snowflake-o"></i>';
            icon = incl.GetSvg_Schneeflocke(0, 4, C.INFOTXTSIZE, C.INFOTXTSIZE);
        }

        let wind = Math.max(element.wind_speed, element.wind_gust);
        if (wind && (wind >= 60)) {
            txt = Math.round(wind);
            //            icon = '<i class="wi wi-darksky-wind"></i>';
            icon = incl.GetSvg_Wind(0, 2, C.INFOTXTSIZE, C.INFOTXTSIZE, element.wind_deg, wind);
            //            icon += incl.GetSvg_Schneeflocke(0, 4, C.INFOTXTSIZE, C.INFOTXTSIZE) ;

        }
        return [icon, txt];
    }

    function GetTempTxt(temp, unit = "") {
        let tempTxt;
        if (typeof temp == 'object') {
            tempTxt = temp.min.toFixed(0) + "/" + temp.max.toFixed(0);
            temp = temp.max;
        }
        else
            tempTxt = temp.toFixed(0);

        if (unit)
            tempTxt += "&nbsp;" + unit

        let s = incl.GetSvg_Thermo(0, 0, 16, 20, temp);
        s += `${tempTxt}`;
        return s;
    }

    function GetWetterTxt(incl, size, cur, Dps, timeZone, noSub = false) {
        const trenn = " &nbsp;";
        const space = "&nbsp;";
        const bewAchtel = Math.round(cur.clouds * 8 / 100);

        const bigger = size + 4;
        let uvi = Math.round(cur.uvi);

        let desc = `<div flex style="font-size:${size}px;margin:left;padding-left:5px;padding-top:0px">`;
        // if (cur.description)
        if (cur.ww) {
            const mx = Mosmix.GetWW_Icon(cur.ww);
            if (mx) {
                desc += `<i class="ww"><svg xmlns="http://www.w3.org/2000/svg" version="1.1" width="${size}" height="${bigger}" viewBox="-25 -25 50 50" transform="translate(0,5)">${mx.svg}</svg></i>` + space;
            }
        }
        if (cur.description)
            desc += cur.description;
        desc += "<br>";
        if (!noSub) {
            const sun = Sun_cl.makeSunTxt(cur, timeZone, 0, incl);
            if (sun) {
                if (sun.sunUpDown)
                    desc += sun.sunUpDown + trenn;
            }
            if (!noSub && cur.Moon && cur.Moon.wi) {
                // desc += `<i class="wi wi-moon-${cur.Moon.wi}-dark"></i>`;
                desc += incl.GetSvg_Mond(0, 1, size - 4, size - 4, cur.Moon); // + trenn + cur.Moon.proz;
            }
            desc += "<br>";
        }
        desc += GetTempTxt(cur.temp, Dps.temp.unit) + trenn;
        if (cur.humidity || cur.dew_point) {
            desc += incl.GetSvg_Humidity(-4, 0, bigger, bigger); // + space;
            if (cur.humidity) {
                desc += cur.humidity.toFixed(0) + space + Dps.humidity.unit;
            }
            if (cur.dew_point) {
                if (cur.humidity)
                    desc += space + "/" + space;
                desc += cur.dew_point.toFixed(0) + space + Dps.dew_point.unit;
            }
            // desc += trenn;
        }
        // desc += "<br>";
        desc += trenn;

        // desc += `<i class="wi wi-darksky-wind"></i>&nbsp;&nbsp;`;
        // desc += `<i class="wi wi-arrow wi-rotate-${cardinal}" style="color:LightSkyBlue; font-size: ${bigger}px;"></i>`;
        let Txt_wind_speed = cur.wind_speed.toFixed();
        if (cur.wind_gust && (cur.wind_gust > cur.wind_speed * 1.1))
            Txt_wind_speed += "/" + cur.wind_gust.toFixed();
        Txt_wind_speed += space + Dps.wind_speed.unit;
        //        console.log(Txt_wind_speed);
        desc += incl.GetSvg_Wind(0, 4, bigger, bigger, cur.wind_deg, Math.max(cur.wind_speed, cur.wind_gust));
        desc += Txt_wind_speed + trenn;
        desc += "<br>";

        const regenIcon = incl.GetSvg_Tropfen(0, -2, C.INFOTXTSIZE, C.INFOTXTSIZE);

        if (cur.rain < 0)
            console.log("rain?", cur.rain);
        let percTxt = "";
        if (cur.precipitation_perc && (Math.round(cur.precipitation_perc) >= 10))
            percTxt = cur.precipitation_perc.toFixed(0) + space + Dps.precipitation_perc.unit + trenn;

        if (cur.precipitation && (Math.round(cur.precipitation) > 0)) {
            desc += regenIcon + space + percTxt + cur.precipitation.toFixed(0) + space + Dps.precipitation.unit + trenn;
            percTxt = "";
        }
        if (cur.snow && (Math.round(cur.snow) > 0)) {
            desc += incl.GetSvg_Schneeflocke(0, 4, C.INFOTXTSIZE, C.INFOTXTSIZE) + space + percTxt + cur.snow.toFixed(0) + space + Dps.snow.unit + trenn;
            percTxt = "";
        }

        if (percTxt)
            desc += regenIcon + space + percTxt;

        desc += incl.GetSvg_WMO2700N(bewAchtel, -2, 2, size, size) + space + bewAchtel + "/8" + trenn;

        if (cur.pressure > 0)
            desc += `<i class="fa fa-compress"></i>&nbsp;` + cur.pressure.toFixed(0) + space + Dps.pressure.unit + trenn;
        if (cur.visibility > 0)
            desc += `<i class="fa fa-eye"></i>&nbsp;` + (cur.visibility / 1000).toFixed(1) + space + "k" + Dps.visibility.unit + trenn;
        if (uvi)
            desc += `<i class="fa fa-sun-o"></i>&nbsp;` + "uvi" + space + uvi + trenn;

        desc += "</div>";
        return desc;
    }

    function getIcon(element, size, noSub) {
        if (element && element.weather && element.weather[0] && element.weather[0].icon) {
            if (!element.description)
                element.description = '<i class="wi wi-owm-' + element.weather[0].icon + '"></i>' + " " + element.weather[0].description; // + " " + element.weather[0].id;
        }

        function WetterIcon(size, data, pay, noSub) {
            const Dps = pay.Dps;
            const timeZone = pay.timeZone;

            if (!size)
                Logg.error(size);
            // let sizeInfo = Math.round(size / 4);
            // let infoPos = size - sizeInfo - 4;

            let icon = incl.GetSvg_Wetter(0, 0, size - 3, size - 3, wDataToStufe(data, pay));
            let border = "border: 1px solid #73AD21;";
            border = "";
            let both = `<div class="tooltip" style="position:relative;height:${size}px;width:${size}px;${border};margin: auto">${icon}`;
            // if (mx) {
            // both += `<div class="ww" style="position: absolute; left: 0px; top: ${infoPos}px;"><svg xmlns="http://www.w3.org/2000/svg" version="1.1" width="${sizeInfo}" height="${sizeInfo}" viewBox="-25 -25 50 50">${mx.svg}</svg></div>`;
            // both += `<div class="ww" style="position: absolute; right: 0px; top: 0px;">${WMO}</div>`;
            // }
            if (Dps && timeZone) {
                const desc = GetWetterTxt(incl, C.INFOTXTSIZE, data, Dps, timeZone, noSub);
                both += `<span class="tooltiptext">${desc}</span>`;
                both += `</div>`;
                // both +=`<p>${pay.ww} - ${mx.desc}</p><p>   - ${mx.descD}</p>`;
            }
            return both;
        }

        return WetterIcon(size, element, pay, noSub);
    }

    let txt = GetWetterTxt(incl, C.INFOTXTSIZE, pay.current, pay.Dps, pay.timeZone, true);
    let back = "background-color:white; ";
    back = "";
    function Uhrzeit(UNIX_timestamp) {
        let dateObject = new Date(UNIX_timestamp * UNIX_MUL);
        return dateObject.toLocaleString(Language, { timeZone: pay.timeZone, hour12: C.Hour12, hour: "numeric", minute: "2-digit" }).toLowerCase();
    }
    let iii = "";
    {
        let descI = "Stand: " + Uhrzeit(pay.update.Akt);
        descI += " Next: " + Uhrzeit(pay.update.Next);
        const iC = incl.GetSvg_Wetter(0, 0, 90, 90, wDataToStufe(pay.current, pay));
        iii += `<span class="tooltip" style="padding:0px;left:0px;top:0px;">${iC}`;
        iii += `<span class="tooltiptext">${descI}</span>`;
        iii += "</span>";
    }
    let icon = `<script>var value = "refresh";this.scope.action = function() { return value; };</script><md-buttonx style="${back} position: relative; margin: auto; height: 90px; width: 90px; padding: 0px; padding: 0px; left: 0px; top: 0px;" ng-click="send({payload:action()})">${iii}</md-button>`;

    let cur = pay.current;
    // prepare forecast data for CSS based ui widget
    let hourly_ri = "";
    const flex = `display:flex;justify-content:center;align-items:top;`; // ;background-color:grey;
    const flexSpan = `width: 100%;text-align:center;`;
    // const flexSpanBot = "width: 100%;text-align:center;justify-content: center;align-items:center;margin-bottom: 0px;";
    const colorTime = "grey"; // "#097479";
    const sysDate = new Date();
    {
        let anz = 0;
        const keys = Object.keys(pay.hourly);
        let start = 0;
        for (let id = 0; id < keys.length; id++) {
            element = pay.hourly[keys[id]];
            if (element.dt * UNIX_MUL < sysDate.getTime())
                continue;
            start = id;
            break;
        }
        for (let id = start; id < keys.length; id += 3)
        // pay.hourly.forEach(function (element, id)
        {
            element = pay.hourly[keys[id]];
            if (anz >= 7)
                break;
            anz++;
            const Sun = new Sun_cl(element.dt, pay.lat, pay.lon);
            element.IsTag = Sun.IsTag;
            // console.log (element.IsTag, timeShort(element.dt, pay.timeZone, true) );
            element.Moon = Moon_Cl.get(pay.lat, new Date(element.dt * UNIX_MUL)).Moon;

            let info = info2(element);
            // info[0] = "aaa";
            hourly_ri += `<div style="${flexSpan}color:${colorTime}">` + timeShort(element.dt, pay.timeZone, true) + getIcon(element, 65, true)
                + `<div style="color:white">` + GetTempTxt(element.temp) + '</div>'
                + '<div style="color:white;text-align:center">' + info[0] + info[1] + '</></div>'
                + '</div>';

        }
    }
    // debugger;
    let daily_ri = "";
    if (pay.daily) {
        let anz = 0;
        pay.daily.forEach(function (element, id) {
            if (TagImJahr(new Date(element.dt * UNIX_MUL,)) == TagImJahr(sysDate)) {
                // console.log ( from + " del " + TagImJahr(new Date(element.dt * UNIX_MUL,) ) +" " +  TagImJahr(sysDate) + " " + dayName(element.dt, pay.timeZone) + " " + dayName(sysDate.getTime()/UNIX_MUL, pay.timeZone))
                return;
            }
            if (anz >= 6)
                return;
            // console.log ( from + " add " + pay.timeZone + " " + new Date(element.dt * UNIX_MUL).toDateString() + " " + TagImJahr(new Date(element.dt * UNIX_MUL) ) +" " +  TagImJahr(sysDate) + " " + dayName(element.dt, pay.timeZone) + " " + dayName(sysDate.getTime()/UNIX_MUL, pay.timeZone))
            anz++;
            element.Moon = Moon_Cl.get(pay.lat, new Date(element.dt * UNIX_MUL)).Moon;
            let info = info2(element);
            daily_ri += `<div style="${flexSpan}color:${colorTime}">` + dayName(element.dt, pay.timeZone)
                + getIcon(element, 65)
                + `<div style="color:white">` + GetTempTxt(element.temp) + '</div>'
                + '<div style="color:white;text-align:center">' + info[0] + info[1] + '</></div>'
                + '</div>';
        });
    }
    /*    .tooltip .tooltiptext::after {
            content: "";
            position: fixed;
            top: 100%;
            left: 50%;
            margin-left: 5px;
            border-width: 5px;
            border-style: solid;
            border-color: transparent transparent transparent transparent;
        }*/

    const mIcon = incl.GetSvg_Mond(0, 0, 20, 20, cur.Moon);
    let desc = cur.Moon.name;
    if (Language == 'de')
        desc = cur.Moon.nameD;
    let mondIcon = `<span class="tooltip" style="padding:0px;left:0px;top:0px;">${mIcon}`;
    mondIcon += `<span class="tooltiptext">${desc}</span>`;
    mondIcon += "</span>";

    let UTC = timeZoneNameShort(pay.timeZone);
    let sunUpDown = Sun_cl.addsunUpDown(pay.Txt, incl);
    if (!sunUpDown)
        sunUpDown = "";
    // border-block-start: 1px solid grey;
    let out = `
<!DOCTYPE html>
<head>
<title>${from}</title>
</head>
<html>
<style>
.tooltip:hover .tooltiptext {
    visibility: visible;
    opacity: 1;
}
.tooltip .tooltiptext {
        font-size: ${C.INFOTXTSIZE}px;
        visibility: hidden;
        background-color: #555;
        color: #fff;
        text-align: left;
        border-radius: 2px;
        padding: 2px 2px;
        position: fixed;
        z-index: 10;
        margin-left: 20px;
        opacity: 0;
        transition: opacity 0.3s;
    }
</style>
<body>
<script>${Rwi};</script>
<div style="width: 498px; padding-top: 0px; justify-content: center; align-items: center;background-color:#424242;color: white;">
<div style="display:flex;justify-content:left;padding:5px; padding-bottom: 0px;">
    <div style="left: 0px; width: 274px; ">${pay.Bez} ${UTC}</div>
    <div style="width: 200px; ">${sunUpDown}</div>
    <div style="right: 24px; width: 24px; ">${mondIcon}</div>
</div>
<div style="display:flex;justify-content:center;align-items:center; padding:0px;">
    <div style="left: 0px; width: 97px; height: 97px;">${icon}</div>
    <div style="left: 98px; width: 391px;">${txt}</div>
</div>
<div>
    <div style="${flex}padding:0px; padding-top: 10px;font-size: ${C.INFOTXTSIZE}px;">
        ${hourly_ri}
    </div>
    <br>
</div>
<div>
    <div style="${flex}padding:0px; padding-top: 10px;font-size: ${C.INFOTXTSIZE}px;">
        ${daily_ri}
    </div>
</div>
</div>
</body>
</html>`;

    let size = 90;
    const iconAkt = Rwi.GetSvg_Wetter(0, 0, size, size, wDataToStufe(pay.current, pay));
    let txt2 = GetWetterTxt(Rwi, C.INFOTXTSIZE, pay.current, pay.Dps, pay.timeZone, false);

    let akt = `
<!DOCTYPE html>
<head>
<title>${from}</title>
</head>
<html>
<body>
<div style="display:flex;justify-content:left;align-items:center; padding:0px;">
    <div style="left: 0px; width: ${size}px; height: ${size}px;">${iconAkt}</div>
    <div style="left: ${size}px;">${txt2}</div>
</div>
</body>
</html>`;
    return { akt: akt, all: out };
}
/*
<div layout="rowInfoIcon" layout-align="space-around start" style="font-size:100%;padding-top: 10px;padding-bottom: 5px">
${hourly_riInfo}
</div>
    <div layout="rowInfoIcon" layout-align="space-around start" style="font-size:100%;padding-top: 5px;padding-bottom: 5px">
        ${daily_riInfo}
    </div>
*/
function wDataToStufe(data, pay) {
    let Stufe = {};
    if (data.ww) {
        Stufe = Mosmix.GetWW_Stufe(data.ww);
        if (Stufe.SchneeRegen_m2 > 0) {
            let temp = data.temp;
            if (data.temp.min)
                temp = data.temp.min;
            if (temp > 5) {
                Stufe.Regen_m3 = Stufe.SchneeRegen_m2;
                Stufe.SchneeRegen_m2 = 0;
            }
        }
    }
    else {
        if (data.weather && data.weather[0] && data.weather[0].id)
            Stufe = idToStufe(data.weather[0].id);
    }

    if (!pay.current || !pay.current.Moon) {
        console.log(pay.current);
        return;
    }
    Stufe.Moon = data.Moon;
    if (!Stufe.Moon)
        Stufe.Moon = pay.current.Moon;

    Stufe.wind = Math.max(data.wind_gust, data.wind_speed);

    Stufe.IsTag = data.IsTag;
    Stufe.Achtel_m8 = Math.round(data.clouds * 8 / 100);
    /*    if (!Stufe.Wolke_m4 && data) {
            switch (Stufe.Achtel_m8) {
                case 1:
                case 2:
                    Stufe.Wolke_m4 = 1;
                    break;
                case 3:
                case 4:
                case 5:
                    Stufe.Wolke_m4 = 2;
                    break;
                case 6:
                case 7:
                    Stufe.Wolke_m4 = 3;
                    break;
                case 8:
                    Stufe.Wolke_m4 = 4;
                    break;
                case 0:
                default:
                    Stufe.Wolke_m4 = 0;
                    break;
            }
        }*/
    return Stufe;
}

/**
* @param {number} deg
*/
function degreesToCardinal(deg) {
    if (deg > 11.25 && deg <= 33.75) { return "NNE"; }
    else if (deg > 33.75 && deg < 56.25) { return "NE"; }
    else if (deg > 56.25 && deg < 78.75) { return "ENE"; }
    else if (deg > 78.75 && deg < 101.25) { return "E"; }
    else if (deg > 101.25 && deg < 123.75) { return "ESE"; }
    else if (deg > 123.75 && deg < 146.25) { return "SE"; }
    else if (deg > 146.25 && deg < 168.75) { return "SSE"; }
    else if (deg > 168.75 && deg < 191.25) { return "S"; }
    else if (deg > 191.25 && deg < 213.75) { return "SSW"; }
    else if (deg > 213.75 && deg < 236.25) { return "SW"; }
    else if (deg > 236.25 && deg < 258.75) { return "WSW"; }
    else if (deg > 258.75 && deg < 281.25) { return "W"; }
    else if (deg > 281.25 && deg < 303.75) { return "WNW"; }
    else if (deg > 303.75 && deg < 326.25) { return "NW"; }
    else if (deg > 326.25 && deg < 348.75) { return "NNW"; }
    else { return "N"; }
}

class Moon_Cl {
    static phases = ['new-moon', 'waxing-crescent-moon', 'quarter-moon', 'waxing-gibbous-moon', 'full-moon', 'waning-gibbous-moon', 'last-quarter-moon', 'waning-crescent-moon'];
    static phasesD = ['Neumond', 'zunehmender Halbmond', 'zunehmender Viertelmond', 'zunehmender Mond', 'Vollmond', 'abnehmender Mond', 'abnehmender Viertelmond', 'abnehmender Halbmond'];
    static wi = ['new', 'wax-cres', 'first-quart', 'wax-gibb', 'full', 'wan-gibb', 'third-quart', 'wan-cres'];
    constructor() {
    }
    static get(lat = null, date = new Date()) {
        let year = date.getFullYear();
        let month = date.getMonth();
        let day = date.getDate();

        let c = 0;
        let e = 0;
        let proz = 0;
        let phase = 0;

        if (month < 3) {
            year--;
            month += 12;
        }

        ++month;
        c = 365.25 * year;
        e = 30.6 * month;
        proz = c + e + day - 694039.09; // jd is total days elapsed
        proz /= 29.5305882; // divide by the moon cycle
        phase = parseInt(proz); // int(jd) -> b, take integer part of jd
        proz -= phase; // subtract integer part to leave fractional part of original jd
        phase = Math.round(proz * 8); // scale fraction from 0-8 and round

        if (phase >= 8) phase = 0; // 0 and 8 are the same so turn 8 into 0
        proz = Math.round(proz * 100);
        return { Moon: { phase: phase, proz: proz, name: this.phases[phase], nameD: this.phasesD[phase], wi: this.wi[phase], lat: lat } };
    }
};

function idToStufe(id) {
    Stufe = {};
    // { IsTag: undefined, wind: 0, Sandsturm_m3: 0, Schauer_m3: 0, Regen_m3: 0, SchneeRegen_m2: 0, Schnee_m3: 0, Niesel_m3: 0, Glaette_b: false, Nebel_b: false, Blitz_b: false, Hagel_b: false }

    switch (id) {
        //Group 3xx: Thunderstorm
        case 200: // thunderstorm with light rain 	11d
            Stufe.Blitz_b = true;
            Stufe.Regen_m3 = 1;
            break;
        case 201: // thunderstorm with rain 	11d
            Stufe.Blitz_b = true;
            Stufe.Regen_m3 = 2;
            break;
        case 202: // thunderstorm with heavy rain 	11d
            Stufe.Blitz_b = true;
            Stufe.Regen_m3 = 3;
            break;
        case 210: // light thunderstorm 	11d
            Stufe.Blitz_b = true;
            break;
        case 211: // thunderstorm 	11d
            Stufe.Blitz_b = true;
            break;
        case 212: // heavy thunderstorm 	11d
            Stufe.Blitz_b = true;
            break;
        case 221: // ragged thunderstorm 	11d
            Stufe.Blitz_b = true;
            break;
        case 230: // thunderstorm with light drizzle 	11d
            Stufe.Blitz_b = true;
            Stufe.Niesel_m3 = 1;
            break;
        case 231: // thunderstorm with drizzle 	11d
            Stufe.Blitz_b = true;
            Stufe.Niesel_m3 = 2;
            break;
        case 232: // thunderstorm with heavy drizzle 	11d
            Stufe.Blitz_b = true;
            Stufe.Niesel_m3 = 3;
            break;
        //Group 3xx: Drizzle
        case 300: // light intensity drizzle 	09d
            Stufe.Niesel_m3 = 1;
            break;
        case 301: // drizzle 	09d
            Stufe.Niesel_m3 = 2;
            break;
        case 302: // heavy intensity drizzle 	09d
            Stufe.Niesel_m3 = 3;
            break;
        case 310: // light intensity drizzle rain 	09d
        case 311: // drizzle rain 	09d
            Stufe.Niesel_m3 = 1;
            Stufe.Regen_m3 = 1;
            break;
        case 312: // heavy intensity drizzle rain 	09d
            Stufe.Niesel_m3 = 2;
            Stufe.Regen_m3 = 2;
            break;
        case 313: // shower rain and drizzle 	09d
            Stufe.Niesel_m3 = 1;
            Stufe.Schauer_m3 = 1;
            break;
        case 314: // heavy shower rain and drizzle 	09d
            Stufe.Niesel_m3 = 3;
            Stufe.Schauer_m3 = 3;
            break;
        case 321: // shower drizzle 	09d
            Stufe.Niesel_m3 = 1;
            break;
        // Group 5xx: Rain
        case 500: // light rain 	10d
            Stufe.Regen_m3 = 1;
            break;
        case 501: // moderate rain 	10d
            Stufe.Regen_m3 = 2;
            break;
        case 502: // heavy intensity rain 	10d
            Stufe.Regen_m3 = 3;
            break;
        case 503: // very heavy rain 	10d
            Stufe.Regen_m3 = 3;
            break;
        case 504: // extreme rain 	10d
            Stufe.Regen_m3 = 3;
            break;
        case 511: // freezing rain 	13d
            Stufe.Glaette_b = true;
            Stufe.Regen_m3 = 3;
            break;
        case 520: // light intensity shower rain 	09d
            Stufe.Schauer_m3 = 1;
            break;
        case 521: // shower rain 	09d
            Stufe.Schauer_m3 = 2;
            break;
        case 522: // heavy intensity shower rain 	09d
            Stufe.Schauer_m3 = 3;
            break;
        case 531: // ragged shower rain 	09d
            Stufe.Schauer_m3 = 3;
            break;
        //Group 6xx: Snow
        case 600: // light snow 	13d
            Stufe.Schnee_m3 = 1;
            break;
        case 601: // snow 	13d
            Stufe.Schnee_m3 = 2;
            break;
        case 602: // heavy snow 	13d
            Stufe.Schnee_m3 = 3;
            break;
        case 611: // sleet 	13d
            Stufe.SchneeRegen_m2 = 1;
            // Stufe.Glaette_b = true;
            break;
        case 612: // light shower sleet 	13d
            Stufe.Schnee_m3 = 1;
            // Stufe.Glaette_b = true;
            break;
        case 613: // shower sleet 	13d
            Stufe.Schnee_m3 = 2;
            // Stufe.Glaette_b = true;
            break;
        case 615: // light rain and snow 	13d
            Stufe.SchneeRegen_m2 = 1;
            // Stufe.Glaette_b = true;
            break;
        case 616: // rain and snow 	13d
            Stufe.SchneeRegen_m2 = 2;
            // Stufe.Glaette_b = true;
            break;
        case 620: // light shower snow 	13d
            Stufe.SchneeRegen_m2 = 1;
            // Stufe.Glaette_b = true;
            break;
        case 621: // shower snow 	13d
            Stufe.SchneeRegen_m2 = 2;
            // Stufe.Glaette_b = true;
            break;
        case 622: // heavy shower snow 	13d
            Stufe.SchneeRegen_m2 = 2;
            // Stufe.Glaette_b = true;
            break;
        // Group 7xx: Atmosphere
        case 701: // mist 	50d
            Stufe.Nebel_b = true;
            break;
        case 711: // smoke 	50d
            Stufe.Nebel_b = true;
            break;
        case 721: // haze 	50d
            Stufe.Hagel_b = true;
            break;
        case 731: // sand/dust whirls 	50d
            Stufe.Sandsturm_m3 = 2;
            break;
        case 741: // fog 	50d
            Stufe.Nebel_b = true;
            break;
        case 751: // sand 	50d
            Stufe.Sandsturm_m3 = 2;
            break;
        case 761: // dust 	50d
            Stufe.Sandsturm_m3 = 2;
            break;
        case 762: // volcanic ash 	50d
            Stufe.Sandsturm_m3 = 2;
            break;
        case 771: // squalls 	50d
            Stufe.wind = 80; // Stufe 4
            break;
        case 781: // tornado 	50d
            break;
        // Group 800: Clear
        case 800: // Clear 	clear sky 	01d
            break;
        // Group 80x: Clouds
        case 801: // few clouds: 11-25% 	02d
            // Stufe.Wolke_m4 = 1;
            break;
        case 802: // scattered clouds: 25-50% 	03d
            // Stufe.Wolke_m4 = 2;
            break;
        case 803: // broken clouds: 51-84% 	04d
            // Stufe.Wolke_m4 = 3;
            break;
        case 804: // overcast clouds: 85-100% 	04d
            // Stufe.Wolke_m4 = 4;
            break;

        default:
            Logg.error("OW Id? " + id);
            break;
    }
    return Stufe;
}
